import unittest

from peyes._utils.event_utils import *
from peyes._utils.pixel_utils import visual_angle_to_pixels
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.Event import BaseEvent


class TestEventUtils(unittest.TestCase):

    def test_calculate_sampling_rate(self):
        t = np.arange(10)
        self.assertEqual(calculate_sampling_rate(t), cnst.MILLISECONDS_PER_SECOND)
        t = np.arange(0, 11, 2)
        self.assertEqual(calculate_sampling_rate(t), cnst.MILLISECONDS_PER_SECOND / 2)
        t = np.arange(10) * cnst.MILLISECONDS_PER_SECOND
        self.assertEqual(calculate_sampling_rate(t), 1)
        t = np.setdiff1d(np.arange(10), [2, 5, 8])
        self.assertEqual(calculate_sampling_rate(t), round(cnst.MILLISECONDS_PER_SECOND / 1.5, 1))

    def test_calculate_num_samples(self):
        self.assertEqual(calculate_num_samples(0, 1000, 1000), 1000)
        self.assertEqual(calculate_num_samples(0, 1000.1, 1000), 1001)
        self.assertEqual(calculate_num_samples(500, 1000, 600), 300)
        self.assertEqual(calculate_num_samples(500, 1000, 600, 400), 400)
        self.assertRaises(ValueError, calculate_num_samples, 0, 1000, 0)
        self.assertRaises(ValueError, calculate_num_samples, 0, 1000, np.inf)
        self.assertRaises(ValueError, calculate_num_samples, np.nan, 1000, 100)

    def test_parse_label(self):
        self.assertEqual(parse_label(EventLabelEnum.FIXATION), EventLabelEnum.FIXATION)
        self.assertEqual(parse_label(BaseEvent.make(EventLabelEnum.FIXATION, np.arange(10))), EventLabelEnum.FIXATION)
        self.assertEqual(parse_label(1), EventLabelEnum.FIXATION)
        self.assertEqual(parse_label("fixation"), EventLabelEnum.FIXATION)
        self.assertEqual(parse_label(1.0), EventLabelEnum.FIXATION)
        self.assertRaises(ValueError, parse_label, 1.5, safe=False)
        self.assertEqual(parse_label(1.5, safe=True), EventLabelEnum.UNDEFINED)
        self.assertRaises(KeyError, parse_label, "foo", safe=False)
        self.assertEqual(parse_label("foo", safe=True), EventLabelEnum.UNDEFINED)
        self.assertRaises(TypeError, parse_label, None, safe=False)
        self.assertEqual(parse_label(None, safe=True), EventLabelEnum.UNDEFINED)

    def test_microsaccade_ratio(self):
        viewer_distance, pixel_size = 60, 0.3
        events = []
        for i in range(5):
            last_x = visual_angle_to_pixels((i+1)/2, viewer_distance, pixel_size)
            sac = BaseEvent.make(
                EventLabelEnum.SACCADE,
                t=np.arange(10),
                x=np.linspace(0, last_x, 10),
                y=np.zeros(10),
                viewer_distance=viewer_distance,
                pixel_size=pixel_size,
            )
            self.assertTrue(np.isclose(sac.amplitude, (i+1)/2))
            events.append(sac)
        self.assertEqual(microsaccade_ratio(events, 1), 0.2)
        events[2] = events[0]
        self.assertEqual(microsaccade_ratio(events, 1), 0.4)
        self.assertEqual(microsaccade_ratio(events, 0.01), 0)
        self.assertTrue(np.isnan(microsaccade_ratio([], 1)))



import os

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import plotly.io as pio

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, PROCESSED_DATA_DIR
import analysis.statistics._helpers as h

pio.renderers.default = "browser"

######################

OUTPUT_DIR = r'S:\jonathan.nir\ISTRC Scholarship'

LABEL = 1
STIMULUS_TYPE = peyes.constants.IMAGE_STR
GT1, GT2 = "RA", "MN"

DET_COLORMAP: u.COLORMAP_TYPE = {
    "Other Human": "#bab0ac",
    "ivt": u.DEFAULT_DISCRETE_COLORMAP[0],
    "ivvt": u.DEFAULT_DISCRETE_COLORMAP[1],
    "idt": u.DEFAULT_DISCRETE_COLORMAP[2],
    "idvt": u.DEFAULT_DISCRETE_COLORMAP[3],
    "engbert": u.DEFAULT_DISCRETE_COLORMAP[4],
    "nh": u.DEFAULT_DISCRETE_COLORMAP[5],
    "remodnav": u.DEFAULT_DISCRETE_COLORMAP[6],
}

# %%
###################
## Event Counts  ##

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}
event_counts_by_labelers = pd.DataFrame({
    lblr: evnts.apply(lambda e: e.label).value_counts() for lblr, evnts in events_by_labelers.items()
}).fillna(0).T.rename(columns=lambda l: peyes.parse_label(l).name)

del stim_trial_ids, all_events, all_labelers, events_by_labelers

fig1 = px.bar(event_counts_by_labelers, title="Event Counts", text_auto=True)
fig1.update_layout(
    title=None,
    xaxis_title="Labeler",
    yaxis_title="Count",
    # legend_title="Event",
    width=1200, height=600,
    legend=dict(title="Event:", orientation="h", yanchor="top", y=1.07, xanchor="left", x=0.21)
)
fig1.show()

peyes.visualize.save_figure(fig1, "event_counts", OUTPUT_DIR, as_png=True)

# %%
######################
##  Sample Metrics  ##

import analysis.statistics.sample_metrics as sm

sample_metrics = pd.concat([
    sm.load_global_metrics(DATASET_NAME, PROCESSED_DATA_DIR, stimulus_type=STIMULUS_TYPE, metric=None),
    sm.load_sdt(DATASET_NAME, PROCESSED_DATA_DIR, label=LABEL, stimulus_type=STIMULUS_TYPE, metric=None)
], axis=0).loc[[peyes.constants.BALANCED_ACCURACY_STR, peyes.constants.COHENS_KAPPA_STR, peyes.constants.D_PRIME_STR]]
fig2 = h.distributions_figure(
    data=sample_metrics, gt1=GT1, gt2=GT2, title="Agreement Metrics (global)", colors=DET_COLORMAP
)
fig2.update_layout(
    yaxis2=dict(range=[0, 1]),
    width=1000, height=500,
    legend=dict(orientation="h", yanchor="bottom", y=-0.25, xanchor="left", x=0.125)
)
fig2.show()

peyes.visualize.save_figure(fig2, "sample_metrics", OUTPUT_DIR, as_png=True)

# %%
######################
## Channel Metrics  ##

import analysis.statistics.channel_sdt as csdt

CHANNEL_TYPES = ["onset", "offset"]

sdt_metrics = csdt.load(
    dataset_name=DATASET_NAME,
    output_dir=PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=STIMULUS_TYPE,
    channel_type=None
).xs(peyes.constants.D_PRIME_STR, level=peyes.constants.METRIC_STR, axis=0)

fig3 = make_subplots(
    rows=2, cols=2, shared_xaxes="all", shared_yaxes="all", row_titles=CHANNEL_TYPES, column_titles=[GT1, GT2],
    vertical_spacing=0.025, horizontal_spacing=0.01
)

for i, gt in enumerate([GT1, GT2]):
    gt_data = sdt_metrics.xs(gt, level=u.GT_STR, axis=1)
    for j, ch_type in enumerate(CHANNEL_TYPES):
        ch_gt_data = gt_data.xs(ch_type, level=peyes.constants.CHANNEL_TYPE_STR, axis=0)
        detectors = u.sort_labelers(ch_gt_data.columns.get_level_values(u.PRED_STR).unique())
        for k, det in enumerate(detectors):
            data = ch_gt_data.xs(det, level=u.PRED_STR, axis=1)
            thresholds = data.index.get_level_values(peyes.constants.THRESHOLD_STR).unique()
            mean = data.mean(axis=1)
            sem = data.std(axis=1) / np.sqrt(data.count(axis=1))
            if det in [GT1, GT2]:
                det_name = "Other Human"
                det_color = "#bab0ac"
                dash = "dot"
            else:
                det_name = det.strip().removesuffix("Detector")
                det_color = u.get_labeler_color(det_name, k, DET_COLORMAP)
                dash = None
            fig3.add_trace(
                row=j + 1, col=i + 1, trace=go.Scatter(
                        x=thresholds, y=mean, error_y=dict(type="data", array=sem),
                        name=det_name, legendgroup=det_name,
                        mode="lines+markers",
                        marker=dict(size=5, color=det_color),
                        line=dict(color=det_color, dash=dash),
                        showlegend=(i == 0 and j == 0),
                    )
            )
        if j == 0:
            fig3.update_yaxes(title_text=r"$d'$", row=i+1, col=1)
        if i == 1:
            fig3.update_xaxes(title_text="Absolute Difference (samples)", row=2, col=j+1)

del i, gt, gt_data, j, ch_type, ch_gt_data, detectors, k, det, data, thresholds, mean, sem, det_name, det_color, dash


fig3.update_layout(
    title="Detection Sensitivity for Varying Differences",
    legend=dict(orientation="h", yanchor="bottom", y=-0.3, xanchor="left", x=0.25),
    width=1400, height=450,
)
fig3.show()

peyes.visualize.save_figure(fig3, "sensitivity_metrics", OUTPUT_DIR, as_png=True)

# %%

dataset = u.load_dataset(DATASET_NAME)
stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)
stim_dataset = dataset[dataset[peyes.constants.TRIAL_ID_STR].isin(stim_trial_ids)]
stim_names = stim_dataset[peyes.constants.STIMULUS_NAME_STR].unique()
subj_ids = stim_dataset[peyes.constants.SUBJECT_ID_STR].unique()
is_na_trial = stim_dataset.groupby(peyes.constants.TRIAL_ID_STR)[[GT1, GT2]].apply(lambda c: c.isna().any())
trial_lengths = stim_dataset.groupby(peyes.constants.TRIAL_ID_STR).apply(lambda t: t[peyes.constants.T].max())



import peyes._utils.constants as cnst
from peyes._DataModels.DatasetLoader import (
    Lund2013DatasetLoader, IRFDatasetLoader, HFCDatasetLoader, GazeComDatasetLoader
)


def get_metadata(dataset_name: str, show: bool = True) -> dict:
    """ Returns the metadata for the specified dataset, and optionally prints the metadata to the console. """
    dataset_name_lower = dataset_name.lower().strip()
    if dataset_name_lower == "lund2013":
        loader_class = Lund2013DatasetLoader
    elif dataset_name_lower == "irf":
        loader_class = IRFDatasetLoader
    elif dataset_name_lower == "hfc":
        loader_class = HFCDatasetLoader
    elif dataset_name_lower == "gazecom":
        loader_class = GazeComDatasetLoader
    else:
        raise NotImplementedError(f"Unknown dataset: {dataset_name}")
    try:
        name = loader_class.name()
    except AttributeError:
        name = ""
    try:
        url = loader_class.url()
    except AttributeError:
        url = ""
    try:
        articles = loader_class.articles()
    except AttributeError:
        articles = []
    try:
        license = loader_class.license()
    except AttributeError:
        license = ""
    metadata = {cnst.NAME_STR: name, cnst.URL_STR: url, cnst.ARTICLES_STR: articles, cnst.LICENSE_STR: license}
    if show:
        print(f"Dataset {cnst.NAME_STR.capitalize()}:\t{name}")
        print(f"{cnst.URL_STR.upper()}:\t{url}")
        print(f"{cnst.LICENSE_STR.capitalize()}:\t{license}")
        print(f"{cnst.ARTICLES_STR.capitalize()}:")
        for article in articles:
            print(f"\t{article}")
    return metadata




from typing import Optional, Union, Sequence, Tuple, Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from charset_normalizer import detect
from plotly.subplots import make_subplots

import peyes

import analysis.utils as u
import analysis.statistics._helpers as h
from peyes._utils.visualization_utils import make_empty_figure
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

###################


def load(
        dataset_name: str,
        output_dir: str,
        label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        stimulus_type: Optional[Union[str, Sequence[str]]] = None,
        matching_schemes: Optional[Union[str, Sequence[str]]] = None,
        metrics: Optional[Union[str, Sequence[str]]] = None,
) -> pd.DataFrame:
    """
    Loads matched-events' SDT metrics and extracts a sub-frame based on the provided parameters.
    Output DataFrame has the following MultiIndex:
    Index:
        level 0: Matching scheme (e.g. window_10, iou_0.5) - matching scheme & threshold
        level 1: Metric (dprime, recall, etc.)
    Columns:
        level 0: Trial ID (1, 2, ...)
        level 1: GT labeler (human annotators, e.g. RA, MN)
        level 2: Pred labeler (detection algorithms, e.g. EngbertDetector, etc.)
    """
    matches_sdt = h.load_data(
        dataset_name=dataset_name, output_dir=output_dir,
        data_dir_name=f"{peyes.constants.MATCHES_STR}_{peyes.constants.METRICS_STR}", label=label,
        filename_suffix="sdt_metrics", iteration=1, stimulus_type=stimulus_type, sub_index=None
    )
    return _extract_sdt_subframe(matches_sdt, matching_schemes, metrics)


def kruskal_wallis_dunns(
        matches_sdt: pd.DataFrame,
        matching_scheme: str,
        gt_cols: Union[str, Sequence[str]],
        metrics: Union[str, Sequence[str]] = None,
        multi_comp: Optional[str] = "fdr_bh",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    sub_frame = _extract_sdt_subframe(matches_sdt, matching_scheme, metrics)
    sub_frame = sub_frame.droplevel(level=[u.MATCHING_SCHEME_STR], axis=0)  # remove single-value levels from index
    statistics, pvalues, dunns, Ns = h.kruskal_wallis_dunns(sub_frame, gt_cols=gt_cols, multi_comp=multi_comp)
    return statistics, pvalues, dunns, Ns


def friedman_nemenyi(
        matches_sdt: pd.DataFrame,
        matching_scheme: str,
        gt_cols: Union[str, Sequence[str]],
        metrics: Union[str, Sequence[str]] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    sub_frame = _extract_sdt_subframe(matches_sdt, matching_scheme, metrics)
    sub_frame = sub_frame.droplevel(level=[u.MATCHING_SCHEME_STR], axis=0)  # remove single-value levels from index
    statistics, pvalues, nemenyi, Ns = h.friedman_nemenyi(sub_frame, gt_cols=gt_cols)
    return statistics, pvalues, nemenyi, Ns


def post_hoc_table(
        ph_data: pd.DataFrame,
        metric: str,
        gt_cols: Union[str, Sequence[str]],
        alpha: float = 0.05,
        marginal_alpha: Optional[float] = 0.075,
) -> pd.DataFrame:
    if isinstance(gt_cols, str):
        gt_cols = [gt_cols]
    return h.create_post_hoc_table(ph_data, metric, *gt_cols, alpha=alpha, marginal_alpha=marginal_alpha)


def single_scheme_figure(
        matches_sdt: pd.DataFrame,
        matching_scheme: str,
        gt1: str,
        metrics: Union[str, Sequence[str]] = None,
        colors: u.COLORMAP_TYPE = None,
        title: str = "",
        gt2: Optional[str] = None,
        only_box: bool = False,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
) -> go.Figure:
    if metrics is None:
        metrics = [
            m for m in matches_sdt.index.get_level_values(peyes.constants.METRIC_STR).unique()
            if m in u.METRICS_CONFIG.keys()
        ]
    sub_frame = _extract_sdt_subframe(matches_sdt, matching_scheme, metrics)
    sub_frame = sub_frame.droplevel(level=[u.MATCHING_SCHEME_STR], axis=0)  # remove single-value levels from index
    if not title:
        split_scheme = matching_scheme.split("_")
        scheme_name, scheme_threshold = "_".join(split_scheme[:-1]), int(split_scheme[-1])
        threshold_str = f"{scheme_threshold:.2f}(a.u.)" if isinstance(scheme_threshold, float) else f"{scheme_threshold} (samples)"
        title = ("Matched Events :: SDT Metrics <br>" +
                 f"<sup>({u.MATCHING_SCHEME_STR.replace("_", " ").title()}: {scheme_name.replace("_", " ").title()}  " +
                 f"Threshold: {threshold_str}</sup>")
    return h.distributions_figure(
        sub_frame, gt1=gt1, gt2=gt2, title=title,
        only_box=only_box, show_other_gt=show_other_gt, share_x=share_x, share_y=share_y
    )


def multi_threshold_figures(
        matches_sdt: pd.DataFrame,
        matching_scheme: str,
        metrics: Union[str, Sequence[str]] = None,
        title: str = "",
        error_bars: Optional[str] = None,
        colors: u.COLORMAP_TYPE = None,
        show_other_gt: bool = True,
        show_err_bands: bool = False,
) -> Dict[str, go.Figure]:
    all_schemes = sorted(
        [ms for ms in matches_sdt.index.get_level_values(u.MATCHING_SCHEME_STR).unique() if ms.startswith(matching_scheme)],
        key=lambda ms: int(ms.split("_")[-1])
    )
    if metrics is None:
        metrics = [
            m for m in matches_sdt.index.get_level_values(peyes.constants.METRIC_STR).unique()
            if m in u.METRICS_CONFIG.keys()
        ]
    subframe = _extract_sdt_subframe(matches_sdt, all_schemes, metrics)
    subframe.sort_index(
        level=subframe.index.names.index(u.MATCHING_SCHEME_STR),
        key=lambda x: [all_schemes.index(s) for s in x],
        axis=0, inplace=True,
    )
    gt_cols = subframe.columns.get_level_values(u.GT_STR).unique()
    figures = dict()
    for gt in gt_cols:
        # create figure for this GT labeler, with subplots for each metric
        gt_subframe = subframe.xs(gt, level=u.GT_STR, axis=1, drop_level=True)
        subframe_metrics = sorted(
            [m for m in gt_subframe.index.get_level_values(peyes.constants.METRIC_STR).unique()],
            key=lambda m: u.METRICS_CONFIG[m][1]
        )
        fig, nrows, ncols = make_empty_figure(
            subtitles=list(map(lambda met: u.METRICS_CONFIG[met][0] if met in u.METRICS_CONFIG else met, subframe_metrics)),
            sharex=False, sharey=False,
        )
        for i, met in enumerate(subframe_metrics):
            r, c = (i, 0) if ncols == 1 else divmod(i, ncols)
            met_frame = gt_subframe.xs(met, level=peyes.constants.METRIC_STR, axis=0, drop_level=True)
            detectors = u.sort_labelers(met_frame.columns.get_level_values(u.PRED_STR).unique())
            for j, det in enumerate(detectors):
                if det not in gt_cols:
                    # current detector is a prediction labeler (detection algorithm)
                    det_name = det.strip().removesuffix("Detector")
                    det_color = u.get_labeler_color(det_name, j, colors)
                    dash = None
                if det in gt_cols and show_other_gt:
                    # current detector is a GT labeler, and we want to refer to it as "Other GT"
                    det_name = "Other GT"
                    det_color = "#bab0ac"
                    dash = "dot"
                met_det_frame = met_frame.xs(det, level=u.PRED_STR, axis=1, drop_level=True)
                # TODO: the following line is the only difference between this func & the similar one in channel_sdt.py,
                #  if we change how the thresholds are stored in the index, we could change this line
                thresholds = met_det_frame.index.to_series(name=peyes.constants.THRESHOLD_STR).apply(lambda ms: int(ms.split("_")[-1]))
                mean = met_det_frame.mean(axis=1)
                errors = h.calc_error_bars(met_det_frame, error_bars)
                fig.add_trace(
                    row=r + 1, col=c + 1, trace=go.Scatter(
                        x=thresholds, y=mean, error_y=dict(type="data", array=errors),
                        name=det_name, legendgroup=det_name,
                        mode="lines+markers",
                        marker=dict(size=5, color=det_color),
                        line=dict(color=det_color, dash=dash),
                        showlegend=i == 0,
                    )
                )
                if show_err_bands and errors is not None and not errors.isna().all():
                    y_upper, y_lower = mean + errors, mean - errors
                    fig.add_trace(
                        row=r + 1, col=c + 1, trace=go.Scatter(
                            x=np.concatenate((thresholds, thresholds[::-1])),
                            y=np.concatenate((y_upper, y_lower[::-1])),
                            fill="toself", fillcolor=det_color, opacity=0.2,
                            line=dict(color=det_color, width=0),
                            name=det_name, legendgroup=det_name, showlegend=False, hoverinfo="skip",
                        )
                    )
            y_range = u.METRICS_CONFIG[met][2] if met in u.METRICS_CONFIG else None
            fig.update_yaxes(row=r + 1, col=c + 1, range=y_range)
        title = title if title else (
                "Matched Events :: SDT Metrics <br>" + f"<sup>(GT: {gt}  " +
                f"{u.MATCHING_SCHEME_STR.replace("_", " ").title()}: {matching_scheme.replace("_", " ").title()})</sup>"
        )
        fig.update_layout(title=title)
        figures[gt] = fig
    return figures


def multi_metric_figure(
        matches_sdt: pd.DataFrame,
        matching_scheme: str,
        metrics: Union[str, Sequence[str]] = None,
        title: str = "",
        error_bars: Optional[str] = None,
        colors: u.COLORMAP_TYPE = None,
        show_other_gt: bool = True,
        show_err_bands: bool = False,
) -> go.Figure:
    all_schemes = sorted(
        [ms for ms in matches_sdt.index.get_level_values(u.MATCHING_SCHEME_STR).unique() if
         ms.startswith(matching_scheme)],
        key=lambda ms: int(ms.split("_")[-1])
    )
    if metrics is None:
        metrics = [
            m for m in matches_sdt.index.get_level_values(peyes.constants.METRIC_STR).unique()
            if m in u.METRICS_CONFIG.keys()
        ]
    subframe = _extract_sdt_subframe(matches_sdt, all_schemes, metrics)
    subframe.sort_index(
        level=subframe.index.names.index(u.MATCHING_SCHEME_STR),
        key=lambda x: [all_schemes.index(s) for s in x],
        axis=0, inplace=True,
    )
    gt_cols = subframe.columns.get_level_values(u.GT_STR).unique()
    assert 0 < len(gt_cols) <= 2, "Only 1 or 2 GT labelers are allowed for multi-channel figure"
    fig = make_subplots(
        rows=len(metrics), cols=len(gt_cols), shared_xaxes=True, shared_yaxes=True,
        vertical_spacing=0.05, horizontal_spacing=0.01,
        row_titles=list(map(lambda met: met.replace("_", " ").title(), metrics)),
        column_titles=list(map(lambda gt: gt.upper(), gt_cols)),
    )
    for r, met in enumerate(metrics):
        for c, gt in enumerate(gt_cols):
            data = subframe.xs(
                met, level=peyes.constants.METRIC_STR, axis=0, drop_level=True).xs(
                gt, level=u.GT_STR, axis=1, drop_level=True
            )
            detectors = u.sort_labelers(data.columns.get_level_values(u.PRED_STR).unique())
            for k, det in enumerate(detectors):
                if det in gt_cols:
                    if show_other_gt:
                        # current detector is a GT labeler, and we want to refer to it as "Other GT"
                        det_name = "Other GT"
                        det_color = "#bab0ac"
                        dash = "dot"
                    else:
                        # current detector is a GT labeler, and we don't want to show it in the figure
                        continue
                else:
                    # current detector is a prediction labeler (detection algorithm)
                    det_name = det.strip().removesuffix("Detector")
                    det_color = u.get_labeler_color(det_name, k, colors)
                    dash = None

                det_data = data.xs(det, level=u.PRED_STR, axis=1)
                # TODO: the following line is the only difference between this func & the similar one in channel_sdt.py,
                #  if we change how the thresholds are stored in the index, we could change this line
                thresholds = det_data.index.to_series(name=peyes.constants.THRESHOLD_STR).apply(lambda ms: int(ms.split("_")[-1]))
                mean = det_data.mean(axis=1)
                errors = h.calc_error_bars(det_data, error_bars)
                fig.add_trace(
                    go.Scatter(
                        x=thresholds, y=mean, error_y=dict(type="data", array=errors),
                        name=det_name, legendgroup=det_name,
                        mode="lines+markers",
                        marker=dict(size=5, color=det_color),
                        line=dict(color=det_color, dash=dash),
                        showlegend=r == 0 and c == 0,
                    ),
                    row=r + 1, col=c + 1
                )
                if show_err_bands and errors is not None and not errors.isna().all():
                    y_upper, y_lower = mean + errors, mean - errors
                    fig.add_trace(
                        go.Scatter(
                            x=np.concatenate((thresholds, thresholds[::-1])),
                            y=np.concatenate((y_upper, y_lower[::-1])),
                            fill="toself", fillcolor=det_color, opacity=0.2,
                            line=dict(color=det_color, width=0),
                            name=det_name, legendgroup=det_name,
                            showlegend=False, hoverinfo="skip",
                        ),
                        row=r + 1, col=c + 1
                    )
            if r == len(metrics) - 1:
                fig.update_xaxes(title_text="Threshold (samples)", row=r + 1, col=c + 1)
            if c == len(gt_cols) - 1:
                # update y-axis range, see https://stackoverflow.com/a/79023935/8543025
                y_range = u.METRICS_CONFIG[met][2] if met in u.METRICS_CONFIG else None
                if y_range:
                    fig.update_yaxes(range=y_range, row=r + 1)
                else:
                    fig.update_yaxes(autorange=True, row=r + 1)
    fig.update_layout(
        title=title if title else (
                "Matched Events :: Metrics for Increasing Thresholds <br>" +
                f"<sup>({u.MATCHING_SCHEME_STR.replace('_', ' ').title()}: " +
                f"{matching_scheme.replace('_', ' ').title()})</sup>"
        ),
        legend=dict(orientation="h", yanchor="top", y=1.045, xanchor="left", x=0.3),
    )
    return fig


def _extract_sdt_subframe(
        matches_sdt: pd.DataFrame,
        matching_schemes: Optional[Union[str, Sequence[str]]] = None,
        metrics: Optional[Union[str, Sequence[str]]] = None,
) -> pd.DataFrame:
    """
    Given a DataFrame of matched-events' SDT metrics that has multi-level index (channel_type, metric, threshold),
    extract a sub-frame based on the provided parameters. Returns a DataFrame with the same index structure as the
    input DataFrame.

    :param matches_sdt: pd.DataFrame; SDT metrics
    :param matching_schemes: str or list of str; name of the matching scheme(s) to extract (e.g. `window_10`, `iou_0.5`)
    :param metrics: str or list of str; metric(s) to extract (e.g. `dprime`, `recall`)

    :return: pd.DataFrame; sub-frame with SDT metrics
    """
    sub_frame = matches_sdt
    if matching_schemes:
        sub_frame = h.extract_subframe(
            sub_frame, level=u.MATCHING_SCHEME_STR, value=matching_schemes, axis=0, drop_single_values=False
        )
    if metrics:
        sub_frame = h.extract_subframe(
            sub_frame, level=peyes.constants.METRIC_STR, value=metrics, axis=0, drop_single_values=False
        )
    return sub_frame



import time

from analysis._article_results.lund2013._helpers import DATASET_NAME, PROCESSED_DATA_DIR, DETECTORS
from analysis.process.full_pipeline import full_pipeline

###################

start = time.time()
print(f"Processing '{DATASET_NAME}' dataset with {len(DETECTORS)} detectors")

print("(1) GLOBAL pipeline")
results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    verbose=False
)
end = time.time()
print(f"\tCompleted GLOBAL pipeline in {end - start:.2f} seconds")

print("(2) FIXATION and SACCADE pipelines")
fix_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    pos_labels=1,
    verbose=False
)
print(f"\tCompleted FIXATION pipeline in {time.time() - end:.2f} seconds")
end = time.time()

print("(3) SACCADE pipeline")
sac_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    pos_labels=2,
    verbose=False
)
print(f"\tCompleted SACCADE pipeline in {time.time() - end:.2f} seconds")
end = time.time()

print(f"Completed full pipeline in {end - start:.2f} seconds")



from typing import Dict, Union

import numpy as np
from tqdm import tqdm

import peyes._utils.constants as cnst
from peyes._DataModels.EventMatcher import OneToOneEventMatchesType


def get_features(
        matches: OneToOneEventMatchesType, *features: str, verbose: bool = False
) -> Union[np.ndarray, Dict[str, np.ndarray]]:
    """
    Calculates the difference in the specified features within each pair of matched events.
    :param matches: a one-to-one mapping from (subset of) ground truth events to (subset_of) predicted events
    :param features: the features to calculate the difference of. Supported features are:
        - 'onset': the difference in start time of the events (ms)
        - 'offset': the difference in end time of the events (ms)
        - 'duration': the difference in duration of the events (ms)
        - 'amplitude': the difference in amplitude of the events (deg)
        - 'azimuth': the difference in azimuth of the events (deg)
        - 'center_pixel_distance': the distance between the center pixels of the events (deg)
        - 'time_overlap': the overlap in time between the events (ms)
        - 'time_iou': the intersection over union in time between the events
        - 'time_l2': the L2 norm of time differences (onset, offset) between the events
    :param verbose: if True, display a progress bar while calculating the features
    :return: the calculated feature differences as a numpy array (for a single feature) or a dictionary of numpy arrays
        (for multiple features).
    """
    results: Dict[str, np.ndarray] = {}
    for feat in tqdm(features, desc="Match Features", disable=not verbose):
        results[feat] = _get_features_impl(matches, feat)
    if len(results) == 1:
        return next(iter(results.values()))
    return results


def _get_features_impl(matches: OneToOneEventMatchesType, feature: str,) -> np.ndarray:
    feature_name = feature.lower().strip().replace(" ", "_").replace("-", "_")
    feature_name = feature_name.removesuffix("s").removesuffix("_difference")   # remove _difference(s)
    if feature_name == "onset":
        return np.array([gt.start_time - pred.start_time for gt, pred in matches.items()])
    if feature_name == "offset":
        return np.array([gt.end_time - pred.end_time for gt, pred in matches.items()])
    if feature_name == cnst.DURATION_STR:
        return np.array([gt.duration - pred.duration for gt, pred in matches.items()])
    if feature_name == cnst.AMPLITUDE_STR:
        return np.array([gt.amplitude - pred.amplitude for gt, pred in matches.items()])
    if feature_name == cnst.AZIMUTH_STR:
        return np.array([gt.azimuth - pred.azimuth for gt, pred in matches.items()])
    if feature_name == "center_pixel_distance":
        return np.array([gt.center_distance(pred) for gt, pred in matches.items()])
    if feature_name == "time_overlap":
        return np.array([gt.time_overlap(pred) for gt, pred in matches.items()])
    if feature_name == "time_iou":
        return np.array([gt.time_iou(pred) for gt, pred in matches.items()])
    if feature_name == "time_l2":
        return np.array([gt.time_l2(pred) for gt, pred in matches.items()])
    raise ValueError(f"Unknown feature: {feature}")



import os

import numpy as np
import plotly.graph_objects as go
import peyes

import analysis.utils as u

DATASET_NAME = "lund2013"
STIMULUS_TYPE = peyes.constants.IMAGE_STR
GT1, GT2 = "RA", "MN"
MULTI_COMP = "bonferroni"   # method for multiple comparisons correction: bonferroni, fdr_bh, holm, etc.
ALPHA, MARGINAL_ALPHA = 0.05, 0.075
NON_500HZ_TRIALS = [33, 34, 39, 44, 54, 58, 63]     # seven of the 63 trials were recorded @ 200Hz instead of 500Hz

_ARTICLE_RESULTS_STR = "article_results"
_FIGURES_STR = "figures"
PROCESSED_DATA_DIR = os.path.join(u.OUTPUT_DIR, _ARTICLE_RESULTS_STR)
FIGURES_DIR = os.path.join(u.OUTPUT_DIR, _ARTICLE_RESULTS_STR, DATASET_NAME, _FIGURES_STR)
os.makedirs(FIGURES_DIR, exist_ok=True)

_default_detector_params = dict(missing_value=np.nan, min_event_duration=4, pad_blinks_time=0)
DETECTORS = {
    "ivt": peyes.create_detector(
        algorithm="ivt", saccade_velocity_threshold=45, **_default_detector_params
    ),
    "ivvt": peyes.create_detector(
        algorithm="ivvt", saccade_velocity_threshold=45, smooth_pursuit_velocity_threshold=26, **_default_detector_params
    ),
    "idt": peyes.create_detector(
        algorithm="idt", dispersion_threshold=2.7, **_default_detector_params
    ),
    # "idt_salvucci": peyes.create_detector(        # Salvucci & Goldberg (2000)
    #     algorithm="idt", dispersion_threshold=1, window_duration=100, **_default_detector_params
    # ),
    "idvt": peyes.create_detector(
        algorithm="idvt", dispersion_threshold=2.7, **_default_detector_params
    ),
    # "idvt_komogortsev": peyes.create_detector(    # Komogortsev & Karpov (2013)
    #     algorithm="idvt",
    #     saccade_velocity_threshold=45,
    #     dispersion_threshold=2,
    #     window_duration=110,
    #     **_default_detector_params
    # ),
    "engbert": peyes.create_detector(
        algorithm="engbert", lambda_param=6, **_default_detector_params
    ),
    "nh": peyes.create_detector(
        algorithm="nh", **_default_detector_params
    ),
    "remodnav": peyes.create_detector(
        algorithm="remodnav", show_warnings=False, **_default_detector_params
    ),
}
for key, detector in DETECTORS.items():
    detector.name = key

## FIGURE CONFIG ##
LABELER_PLOTTING_CONFIG = {
    # labeler -> (order, color, line-style)
    'Other Human': (0, "#bab0ac", 'dot'),
    'RA': (1, u.DEFAULT_DISCRETE_COLORMAP[len(DETECTORS) + 1], 'dot'),
    'MN': (2, u.DEFAULT_DISCRETE_COLORMAP[len(DETECTORS) + 2], 'dot'),
    **{key: (i+2 ,u.DEFAULT_DISCRETE_COLORMAP[i], None) for i, key in enumerate(DETECTORS.keys())}
}


def save_fig(fig: go.Figure, fig_id: int, panel_id: str, panel_name: str, is_supp: bool):
    fig_path = os.path.join(FIGURES_DIR, f"{'supp-' if is_supp else ''}fig{fig_id}")
    os.makedirs(fig_path, exist_ok=True)
    file_name = f"{panel_id}_{panel_name}"
    peyes.visualize.save_figure(fig, file_name, fig_path, as_png=True, as_json=True)



from typing import Optional

import pandas as pd
import sklearn.metrics as met

from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelSequenceType

from peyes._utils.event_utils import parse_label as _parse_label
from peyes._utils.metric_utils import transition_matrix as _transition_matrix

_GROUND_TRUTH_STR = "Ground Truth"
_PREDICTION_STR = "Prediction"


def label_counts(
        seq: UnparsedEventLabelSequenceType
) -> pd.Series:
    """
    Counts the occurrences of each event-label in the given sequence.
    Returns a pandas Series mapping each event-label to its count.
    """
    if seq is None:
        return pd.Series({l: 0 for l in EventLabelEnum})
    seq = [_parse_label(l) for l in seq]
    counts = pd.Series(seq).value_counts()
    return counts.reindex([l for l in EventLabelEnum], fill_value=0)


def transition_matrix(
        seq: UnparsedEventLabelSequenceType,
        normalize_rows: bool = False
) -> pd.DataFrame:
    """
    Calculates the transition matrix from a sequence of event-labels and returns a DataFrame where rows indicate the
    origin label and columns indicate the destination label. If `normalize_rows` is True, the matrix will be normalized
    by the sum of each row, i.e. contains transition **probabilities** and not **counts**.

    Returns a DataFrame where rows indicate the origin label and columns indicate the destination label.
    """
    seq = [_parse_label(l) for l in seq]
    return _transition_matrix(seq, normalize_rows)


def confusion_matrix(
        ground_truth: UnparsedEventLabelSequenceType,
        prediction: UnparsedEventLabelSequenceType,
        labels: Optional[UnparsedEventLabelSequenceType] = None,
) -> pd.DataFrame:
    """
    Calculates the confusion matrix from the given ground-truth and predicted event-labels.
    If `labels` is not None, the matrix will be restricted to the given labels; otherwise it will contain all unique labels.
    """
    ground_truth = [_parse_label(l) for l in ground_truth]
    prediction = [_parse_label(l) for l in prediction]
    labels = list(EventLabelEnum) if labels is None else list(set(_parse_label(l) for l in labels))
    conf = met.confusion_matrix(ground_truth, prediction, labels=labels)
    df = pd.DataFrame(conf, index=labels, columns=labels)
    df.index.name = _GROUND_TRUTH_STR
    df.columns.name = _PREDICTION_STR
    return df



import unittest

import numpy as np

from peyes._utils.vector_utils import *


class TestVectorUtils(unittest.TestCase):

    def test_is_one_dimensional(self):
        self.assertTrue(is_one_dimensional([1, 2, 3]))
        self.assertTrue(is_one_dimensional([[1], [2], [3]]))
        self.assertTrue(is_one_dimensional([[1, 2, 3]]))
        self.assertFalse(is_one_dimensional([[1, 2], [3, 4]]))
        self.assertRaises(ValueError, is_one_dimensional, [[1, 2], [3]])

    def test_pair_boolean_arrays(self):
        arr1 = np.array([True, False, True, False, True])
        arr2 = np.array([False, True, True, False, True])
        obs = pair_boolean_arrays(arr1, arr2)
        exp = np.array([[0, 1], [2, 2], [4, 4]])
        self.assertTrue(np.array_equal(obs, exp))
        arr2 = np.array([False, True, False, False, True])
        obs = pair_boolean_arrays(arr1, arr2)
        exp = np.array([[0, 1], [4, 4]])
        self.assertTrue(np.array_equal(obs, exp))
        arr2 = np.array([True, True, False, False, True])
        obs = pair_boolean_arrays(arr1, arr2)
        exp = np.array([[0, 0], [4, 4]])
        self.assertTrue(np.array_equal(obs, exp))
        arr2 = np.array([False, False, False, False, False])
        obs = pair_boolean_arrays(arr1, arr2)
        exp = np.empty((0, 2))
        self.assertTrue(np.array_equal(obs, exp))
        arr2 = np.array([True, False, True])
        obs = pair_boolean_arrays(arr1, arr2)
        exp = np.array([[0, 0], [2, 2]])
        self.assertTrue(np.array_equal(obs, exp))
        self.assertRaises(ValueError, pair_boolean_arrays, [[True, False], [True, False]], [True, False])
        self.assertRaises(ValueError, pair_boolean_arrays, [True, False], [[True, False], [True, False]])

    def test_get_chunk_indices(self):
        arr = [1, 1, 1, 2, 2, 3, 3, 3, 3]
        obs = get_chunk_indices(arr)
        exp = [np.array([0, 1, 2]), np.array([3, 4]), np.array([5, 6, 7, 8])]
        self.assertTrue(all([np.array_equal(o, e, equal_nan=True) for o, e in zip(obs, exp)]))
        arr[2] = np.nan
        obs = get_chunk_indices(arr)
        exp = [np.array([0, 1]), np.array([2]), np.array([3, 4]), np.array([5, 6, 7, 8])]
        self.assertTrue(all([np.array_equal(o, e, equal_nan=True) for o, e in zip(obs, exp)]))
        arr = np.arange(-5, 5)
        obs = get_chunk_indices(arr)
        exp = [np.array([i]) for i in range(10)]
        self.assertTrue(all([np.array_equal(o, e, equal_nan=True) for o, e in zip(obs, exp)]))

    def test_merge_chunks(self):
        arr = np.array([1, 1, 1, 1, 2, 2, 1, 1, 1])
        self.assertTrue(np.array_equal(merge_chunks(arr, 2), np.ones_like(arr)))
        self.assertTrue(np.array_equal(merge_chunks(arr, 1), arr))
        arr = np.array([1, 1, 1, 1, 2, 2, 3, 3, 3])
        self.assertTrue(np.array_equal(merge_chunks(arr, 2), arr))
        self.assertTrue(np.array_equal(merge_chunks(arr, 1), arr))
        self.assertRaises(ValueError, merge_chunks, np.array([[1, 2], [3, 4]]), 2)
        self.assertRaises(ValueError, merge_chunks, np.array([1, 2, 3]), -1)

    def test_reset_short_chunks(self):
        arr = np.array([1, 1, 1, 1, 2, 2, 1, 1, 1])
        self.assertTrue(np.array_equal(reset_short_chunks(arr, 3, 0), np.array([1, 1, 1, 1, 0, 0, 1, 1, 1])))
        self.assertTrue(np.array_equal(reset_short_chunks(arr, 2, 0), arr))
        arr = np.array([1, 1, 1, 1, 2, 2, 3, 3, 3])
        self.assertTrue(np.array_equal(reset_short_chunks(arr, 3, 0), np.array([1, 1, 1, 1, 0, 0, 3, 3, 3])))
        self.assertTrue(np.array_equal(reset_short_chunks(arr, 2, 0), arr))
        self.assertRaises(ValueError, reset_short_chunks, np.array([[1, 2], [3, 4]]), 2, 0)
        self.assertRaises(ValueError, reset_short_chunks, np.array([1, 2, 3]), -1, 0)



from analysis._parameter_exploration._helpers import DATASET_NAME, PROCESSED_DATA_DIR, DETECTORS
from analysis.process.full_pipeline import full_pipeline

###################

results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    verbose=False
)

fix_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    pos_labels=1,
    verbose=False
)

sac_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DETECTORS.values(),
    num_iterations=1,
    pos_labels=2,
    verbose=False
)



import os
from typing import List, Optional, Union, Tuple

import pandas as pd
from tqdm import tqdm

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType
from peyes.sample_metrics._calculate_metrics import _GLOBAL_METRICS, _SDT_METRICS

import analysis.utils as u
import analysis.process._helpers as h


def run_default(
        dataset_name: str,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
    default_output_dir = h.get_default_output_dir(dataset_name)
    try:
        labels = pd.read_pickle(os.path.join(default_output_dir, f"{peyes.constants.LABELS_STR}.pkl"))
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Couldn't find `{peyes.constants.LABELS_STR}.pkl` in {default_output_dir}. Please preprocess the dataset first."
        )
    sample_metrics_dir = os.path.join(default_output_dir, f"{peyes.constants.SAMPLE_STR}_{peyes.constants.METRICS_STR}")
    os.makedirs(sample_metrics_dir, exist_ok=True)
    if pos_labels is None or len(pos_labels) == 0:
        pos_labels = None
    elif isinstance(pos_labels, UnparsedEventLabelType):
        pos_labels = [pos_labels]

    # calculate SDT metrics, regardless of whether pos_labels is None or not
    sdt_metrics_path = os.path.join(
        sample_metrics_dir, u.get_filename_for_labels(
            pos_labels, suffix=f"{u.SDT_STR}_{peyes.constants.METRICS_STR}", extension="pkl"
        )
    )
    try:
        sdt_metrics = pd.read_pickle(sdt_metrics_path)
    except FileNotFoundError:
        sdt_metric_names = _SDT_METRICS
        if not pos_labels:
            sdt_metric_names -= {peyes.constants.D_PRIME_STR, peyes.constants.CRITERION_STR}
        sdt_metrics = calculate_sdt_sample_metrics(
            labels, u.DATASET_ANNOTATORS[dataset_name], metrics=sdt_metric_names, pos_labels=pos_labels
        )
        sdt_metrics.to_pickle(sdt_metrics_path)
    if pos_labels is not None:
        return sdt_metrics

    # calculate global metrics only if pos_labels is None
    global_metrics_path = os.path.join(
        sample_metrics_dir, u.get_filename_for_labels(
            labels=pos_labels, suffix=f"{u.GLOBAL_STR}_{peyes.constants.METRICS_STR}", extension="pkl"
        )
    )
    try:
        global_metrics = pd.read_pickle(global_metrics_path)
    except FileNotFoundError:
        global_metrics = calculate_global_sample_metrics(
            labels, u.DATASET_ANNOTATORS[dataset_name], metrics=_GLOBAL_METRICS
        )
        global_metrics.to_pickle(global_metrics_path)
    return sdt_metrics, global_metrics


def calculate_global_sample_metrics(
        labels: pd.DataFrame,
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        metrics: List[str] = _GLOBAL_METRICS,
) -> pd.DataFrame:
    pred_labelers = h.check_labelers(labels, pred_labelers)
    metrics = set(m.lower().strip() for m in metrics) or set(_GLOBAL_METRICS)
    unknown_metrics = metrics - set(u.METRICS_CONFIG.keys())
    if unknown_metrics:
        raise ValueError(f"Unknown metrics: {unknown_metrics}")
    non_global_metrics = set(metrics) - set(_GLOBAL_METRICS)
    if non_global_metrics:
        raise ValueError(f"Metrics {non_global_metrics} are not \"global\" metrics.")
    metrics = sorted(metrics, key=lambda m: u.METRICS_CONFIG[m][1])
    return _calculate_impl(
        labels,
        gt_labelers,
        pred_labelers,
        metrics,
        f"Sample Metrics :: Global",
    )


def calculate_sdt_sample_metrics(
        labels: pd.DataFrame,
        gt_labelers: List[str],
        pred_labelers: List[str] = None,
        metrics: List[str] = _SDT_METRICS,
        pos_labels: Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType] = None,
        average: str = "weighted",
        correction: str = "loglinear",
) -> pd.DataFrame:
    pred_labelers = h.check_labelers(labels, pred_labelers)
    metrics = set(m.lower().strip() for m in metrics) or set(_SDT_METRICS)
    unknown_metrics = metrics - set(u.METRICS_CONFIG.keys())
    if unknown_metrics:
        raise ValueError(f"Unknown metrics: {unknown_metrics}")
    non_sdt_metrics = set(metrics) - set(_SDT_METRICS)
    if non_sdt_metrics:
        raise ValueError(f"Metrics {non_sdt_metrics} are not \"SDT\" metrics.")
    if pos_labels is None:
        # cannot calculate d-prime and criterion without specifying a subset of positive labels
        metrics = metrics - {peyes.constants.D_PRIME_STR, peyes.constants.CRITERION_STR}
    metrics = sorted(metrics, key=lambda m: u.METRICS_CONFIG[m][1])
    return _calculate_impl(
        labels,
        gt_labelers,
        pred_labelers,
        metrics,
        f"Sample Metrics :: SDT ({pos_labels})",
        pos_labels,
        average,
        correction
    )


def _calculate_impl(
        labels: pd.DataFrame,
        gt_labelers: List[str],
        pred_labelers: List[str],
        metrics: List[str],
        description: str,
        pos_labels: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        average: str = "weighted",
        correction: str = "loglinear",
) -> pd.DataFrame:
    results = dict()
    for tr in tqdm(labels.columns.get_level_values(peyes.constants.TRIAL_ID_STR).unique(), desc=description):
        for gt_labeler in gt_labelers:
            gt_min_iteration = labels.xs(
                (tr, gt_labeler), axis=1, level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR]
            ).columns.min()
            gt_labels = labels[tr, gt_labeler, gt_min_iteration].dropna().values.flatten()
            if gt_labels.size == 0:
                continue
            for pred_labeler in pred_labelers:
                pred_labels_all_iters = labels.xs((tr, pred_labeler), axis=1,
                                                  level=[peyes.constants.TRIAL_ID_STR, peyes.constants.LABELER_STR])
                for pred_it in pred_labels_all_iters.columns.get_level_values(peyes.constants.ITERATION_STR).unique():
                    if (pred_labeler == gt_labeler) and (pred_it == gt_min_iteration):
                        continue
                    pred_labels = pred_labels_all_iters[pred_it].dropna().values.flatten()
                    if pred_labels.size == 0:
                        continue
                    res = peyes.sample_metrics.calculate(
                        gt_labels, pred_labels, *metrics, pos_labels=pos_labels, average=average, correction=correction
                    )
                    results[(tr, gt_labeler, pred_labeler, pred_it)] = res
    results_df = pd.DataFrame(results).T
    results_df.index.names = [peyes.constants.TRIAL_ID_STR, u.GT_STR, u.PRED_STR, peyes.constants.ITERATION_STR]
    results_df = results_df[sorted(results_df.columns)]
    return results_df.T



from analysis._default_values._helpers import DATASET_NAME, PROCESSED_DATA_DIR, DEFAULTS_DETECTORS
from analysis.process.full_pipeline import full_pipeline

###################

results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DEFAULTS_DETECTORS.values(),
    verbose=False
)

fix_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DEFAULTS_DETECTORS.values(),
    pos_labels=1,
    verbose=False
)

sac_results = full_pipeline(
    output_dir=PROCESSED_DATA_DIR,
    dataset_name=DATASET_NAME,
    detectors=DEFAULTS_DETECTORS.values(),
    pos_labels=2,
    verbose=False
)



import numpy as np
import plotly.graph_objects as go

import peyes._utils.visualization_utils as vis_utils
from peyes._DataModels.EventLabelEnum import EventLabelEnum, EventLabelSequenceType


def scarfplot_comparison_figure(
        t: np.ndarray, *labels, **kwargs
) -> go.Figure:
    """
    Creates a figure with multiple scarfplots stacked on top of each other.
    :param t: The time axis.
    :param labels: multiple sequences of event labels to be plotted, each must have the same length as `t`.
    :keyword scarf_height: The height (in y-axis units) of each scarfplot, default is 1.
    :keyword scarf_spacing: The spacing (in y-axis units) between scarfplots, default is 0.25.
    :keyword label_colors: A dictionary mapping event labels to their respective colors. If a label is missing, the
        default color is used.
    :keyword title: The title of the figure, default is "Scarfplot Comparison".
    :keyword names: The names of the scarfplots, default is [0, 1, 2, ...].
    :return: The figure with the scarfplots.
    """
    num_scarfs = len(labels)
    scarf_height = kwargs.get("scarf_height", 1)
    scarf_spacing = 1 + kwargs.get("scarf_spacing", 0.25)
    label_colors = vis_utils.get_label_colormap(kwargs.get("label_colors", None))
    fig = go.Figure()
    for i, l in enumerate(labels):
        bottom, top = scarf_spacing * i * scarf_height, (scarf_spacing * i + 1) * scarf_height
        fig = add_scarfplot_to_figure(fig, t, l, top, bottom, label_colors=label_colors, show_colorbar=i == 0)
    names = kwargs.get("names", [str(i) for i in range(num_scarfs)])
    assert len(names) == num_scarfs
    fig.update_layout(
        title=kwargs.get("title", "Scarfplot Comparison"),
        yaxis=dict(
            range=[0, 1 + scarf_spacing * scarf_height * (num_scarfs - 1)],
            tickmode='array',
            tickvals=[(scarf_spacing * i + 0.5) * scarf_height for i in range(num_scarfs)],
            ticktext=names,
        ),
    )
    return fig


def add_scarfplot_to_figure(
        fig: go.Figure,
        t: np.ndarray,
        labels: EventLabelSequenceType,
        top: float,
        bottom: float,
        label_colors: vis_utils.LabelColormapType = None,
        colorbar_length: float = 1,
        colorbar_thickness: int = 25,
        show_colorbar: bool = True,
) -> go.Figure:
    """
    Adds a scarfplot to the provided figure.

    :param fig: the figure to add the scarfplot to.
    :param t: the time axis, must have the same length as `labels`.
    :param labels: the event labels to plot, must have the same length as `t`.
    :param top: the top y-coordinate of the scarfplot.
    :param bottom: the bottom y-coordinate of the scarfplot.
    :param label_colors: a dictionary mapping event labels to their respective colors. If a label is missing, the
        default color is used.
    :param colorbar_length: the length of the colorbar (range [0, 1] where 1 is the full height of the plot), default is 1.
    :param colorbar_thickness: the thickness of the colorbar, default is 25.
    :param show_colorbar: whether to show the colorbar, default is True.
    """
    assert len(t) == len(labels), f"Length mismatch: len(t)={len(t)} != {len(labels)}=len(labels)"
    label_colors = vis_utils.get_label_colormap(label_colors)
    label_colors = {k: v for k, v in label_colors.items() if k in labels}   # Remove unused colors
    colormap, tick_centers = _discrete_colormap(label_colors)
    # # Normalize tick centers to [0, 1]
    # see https://community.plotly.com/t/heatmap-colorbar-displays-ticks-in-incorrect-locations/84278/3?u=jonnir
    tick_centers = tick_centers * colorbar_length * (np.max(tick_centers) - np.min(tick_centers)) / len(tick_centers)
    scarfplot = go.Heatmap(
        x=t,
        y=[bottom, top],
        z=[np.asarray(labels, dtype=EventLabelEnum)],
        zmin=min(EventLabelEnum),
        zmax=max(EventLabelEnum),
        colorscale=colormap,
        colorbar=dict(
            len=colorbar_length,
            thickness=colorbar_thickness,
            tickvals=tick_centers,
            ticktext=list(map(lambda lbl: EventLabelEnum(lbl).name, label_colors.keys())),
        ),
        showscale=show_colorbar,
    )
    fig.add_trace(scarfplot)
    return fig


def _discrete_colormap(colors: dict):
    borders = np.arange(len(colors) + 1)
    centers = (borders[1:] + borders[:-1]) / 2
    borders = borders / len(colors)  # Normalize to [0, 1]
    colormap = []
    for i, key in enumerate(sorted(colors.keys())):
        colormap.extend([(borders[i], f"rgb{colors[key]}"), (borders[i + 1], f"rgb{colors[key]}")])
    return colormap, centers



from typing import Optional, Union, Tuple, Sequence, Dict

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

import peyes

import analysis.utils as u
import analysis.statistics._helpers as h
from peyes._utils.visualization_utils import make_empty_figure
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

###################


def load(
        dataset_name: str,
        output_dir: str,
        label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        stimulus_type: Optional[Union[str, Sequence[str]]] = None,
        channel_type: Optional[str] = None,
        threshold: Optional[Union[int, Sequence[int]]] = None,
        metrics: Optional[Union[str, Sequence[str]]] = None,
) -> pd.DataFrame:
    """
    Loads channel SDT metrics and re-arranges data for simpler analysis.
    Output DataFrame has the following MultiIndex:
    Index:
        level 0: Channel type (onset/offset)
        level 1: Metric (recall, precision, etc.)
        level 2: Threshold (0, 1, 2, ...)
    Columns:
        level 0: Trial ID (1, 2, ...)
        level 1: GT labeler (human annotators, e.g. RA, MN)
        level 2: Pred labeler (detection algorithms, e.g. EngbertDetector, etc.)
    """
    sdt_metrics = h.load_data(
        dataset_name=dataset_name, output_dir=output_dir,
        data_dir_name=peyes.constants.SAMPLES_CHANNEL_STR, label=label,
        filename_suffix="sdt_metrics", iteration=1, stimulus_type=stimulus_type, sub_index=None
    )
    sdt_metrics = sdt_metrics.stack(level=peyes.constants.METRIC_STR, future_stack=True)
    sdt_metrics = sdt_metrics.reorder_levels(
        [peyes.constants.CHANNEL_TYPE_STR, peyes.constants.METRIC_STR, peyes.constants.THRESHOLD_STR], axis=0
    )
    return _extract_sdt_subframe(sdt_metrics, channel_type, threshold, metrics)


def wilcoxon(
        sdt_metrics: pd.DataFrame,
        channel_type: str,
        threshold: int,
        gt_cols: Union[str, Sequence[str]],
        metrics: Union[str, Sequence[str]] = None,
        alternative: str = "two-sided",
        method: str = "auto",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    sub_frame = _extract_sdt_subframe(sdt_metrics, channel_type, threshold, metrics)
    sub_frame = sub_frame.droplevel(  # remove single-value levels from index
        level=[peyes.constants.CHANNEL_TYPE_STR, peyes.constants.THRESHOLD_STR], axis=0
    )
    statistics, pvalues, Ns = h.wilcoxon(sub_frame, gt_cols=gt_cols, alternative=alternative, method=method)
    return statistics, pvalues, Ns



def kruskal_wallis_dunns(
        sdt_metrics: pd.DataFrame,
        channel_type: str,
        threshold: int,
        gt_cols: Union[str, Sequence[str]],
        metrics: Union[str, Sequence[str]] = None,
        multi_comp: Optional[str] = "fdr_bh",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    sub_frame = _extract_sdt_subframe(sdt_metrics, channel_type, threshold, metrics)
    sub_frame = sub_frame.droplevel(  # remove single-value levels from index
        level=[peyes.constants.CHANNEL_TYPE_STR, peyes.constants.THRESHOLD_STR], axis=0
    )
    statistics, pvalues, dunns, Ns = h.kruskal_wallis_dunns(sub_frame, gt_cols=gt_cols, multi_comp=multi_comp)
    return statistics, pvalues, dunns, Ns


def friedman_nemenyi(
        sdt_metrics: pd.DataFrame,
        channel_type: str,
        threshold: int,
        gt_cols: Union[str, Sequence[str]],
        metrics: Union[str, Sequence[str]] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    sub_frame = _extract_sdt_subframe(sdt_metrics, channel_type, threshold, metrics)
    sub_frame = sub_frame.droplevel(  # remove single-value levels from index
        level=[peyes.constants.CHANNEL_TYPE_STR, peyes.constants.THRESHOLD_STR], axis=0
    )
    statistics, pvalues, nemenyi, Ns = h.friedman_nemenyi(sub_frame, gt_cols=gt_cols)
    return statistics, pvalues, nemenyi, Ns


def post_hoc_table(
        ph_data: pd.DataFrame,
        metric: str,
        gt_cols: Union[str, Sequence[str]],
        alpha: float = 0.05,
        marginal_alpha: Optional[float] = 0.075,
) -> pd.DataFrame:
    if isinstance(gt_cols, str):
        gt_cols = [gt_cols]
    return h.create_post_hoc_table(ph_data, metric, *gt_cols, alpha=alpha, marginal_alpha=marginal_alpha)


def single_threshold_figure(
        sdt_metrics: pd.DataFrame,
        channel_type: str,
        threshold: int,
        gt1: str,
        metrics: Union[str, Sequence[str]] = None,
        colors: u.COLORMAP_TYPE = None,
        title: str = "",
        gt2: Optional[str] = None,
        only_box: bool = False,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
) -> go.Figure:
    if metrics is None:
        metrics = [
            m for m in sdt_metrics.index.get_level_values(peyes.constants.METRIC_STR).unique()
            if m in u.METRICS_CONFIG.keys()
        ]
    sub_frame = _extract_sdt_subframe(sdt_metrics, channel_type, threshold, metrics)
    sub_frame = sub_frame.droplevel(  # remove single-value levels from index
        level=[peyes.constants.CHANNEL_TYPE_STR, peyes.constants.THRESHOLD_STR], axis=0
    )
    title = title if title else (
        "Samples Channel :: SDT Metrics <br>" +
        f"<sup>(Channel: {channel_type}  Max Difference: {threshold} samples)</sup>"
    )
    fig = h.distributions_figure(
        sub_frame, gt1=gt1, gt2=gt2, colors=colors, title=title,
        only_box=only_box, show_other_gt=show_other_gt, share_x=share_x, share_y=share_y,
    )
    return fig


def multi_threshold_figures(
        sdt_metrics: pd.DataFrame,
        channel_type: str,
        metrics: Union[str, Sequence[str]] = None,
        title: str = "",
        error_bars: Optional[str] = None,
        colors: u.COLORMAP_TYPE = None,
        show_other_gt: bool = True,
        show_err_bands: bool = False,
) -> Dict[str, go.Figure]:
    if metrics is None:
        metrics = [
            m for m in sdt_metrics.index.get_level_values(peyes.constants.METRIC_STR).unique()
            if m in u.METRICS_CONFIG.keys()
        ]
    subframe = _extract_sdt_subframe(sdt_metrics, channel_type, None, metrics)
    subframe = subframe.droplevel(peyes.constants.CHANNEL_TYPE_STR, axis=0)    # remove single-value levels from index
    gt_cols = subframe.columns.get_level_values(u.GT_STR).unique()
    assert 0 < len(gt_cols) <= 2, "Only 1 or 2 GT labelers are allowed for multi-threshold figures"
    figures = dict()
    for gt in gt_cols:
        # create figure for this GT labeler, with subplots for each metric
        gt_subframe = subframe.xs(gt, level=u.GT_STR, axis=1, drop_level=True)
        subframe_metrics = sorted(
            [m for m in gt_subframe.index.get_level_values(peyes.constants.METRIC_STR).unique()],
            key=lambda m: u.METRICS_CONFIG[m][1]
        )
        fig, nrows, ncols = make_empty_figure(
            subtitles=list(map(lambda met: u.METRICS_CONFIG[met][0] if met in u.METRICS_CONFIG else met, subframe_metrics)),
            sharex=False, sharey=False,
        )
        for i, met in enumerate(subframe_metrics):
            r, c = (i, 0) if ncols == 1 else divmod(i, ncols)
            met_frame = gt_subframe.xs(met, level=peyes.constants.METRIC_STR, axis=0, drop_level=True)
            detectors = u.sort_labelers(met_frame.columns.get_level_values(u.PRED_STR).unique())
            for j, det in enumerate(detectors):
                if det not in gt_cols:
                    # current detector is a prediction labeler (detection algorithm)
                    det_name = det.strip().removesuffix("Detector")
                    det_color = u.get_labeler_color(det_name, j, colors)
                    dash = None
                if det in gt_cols and show_other_gt:
                    # current detector is a GT labeler, and we want to refer to it as "Other GT"
                    det_name = "Other GT"
                    det_color = "#bab0ac"
                    dash = "dot"
                met_det_frame = met_frame.xs(det, level=u.PRED_STR, axis=1, drop_level=True)
                thresholds = met_det_frame.index.get_level_values(peyes.constants.THRESHOLD_STR).unique()
                mean = met_det_frame.mean(axis=1)
                errors = h.calc_error_bars(met_det_frame, error_bars)
                fig.add_trace(
                    row=r + 1, col=c + 1, trace=go.Scatter(
                        x=thresholds, y=mean, error_y=dict(type="data", array=errors),
                        name=det_name, legendgroup=det_name,
                        mode="lines+markers",
                        marker=dict(size=5, color=det_color),
                        line=dict(color=det_color, dash=dash),
                        showlegend=i == 0,
                    )
                )
                if show_err_bands and errors is not None and not errors.isna().all():
                    y_upper, y_lower = mean + errors, mean - errors
                    fig.add_trace(
                        row=r + 1, col=c + 1, trace=go.Scatter(
                            x=np.concatenate((thresholds, thresholds[::-1])),
                            y=np.concatenate((y_upper, y_lower[::-1])),
                            fill="toself", fillcolor=det_color, opacity=0.2,
                            line=dict(color=det_color, width=0),
                            name=det_name, legendgroup=det_name, showlegend=False, hoverinfo="skip",
                        )
                    )
            y_range = u.METRICS_CONFIG[met][2] if met in u.METRICS_CONFIG else None
            fig.update_yaxes(row=r + 1, col=c + 1, range=y_range)
        title = title if title else (
                "Samples Channel :: SDT Metrics <br>" + f"<sup>(GT: {gt}  Channel: {channel_type})</sup>"
        )
        fig.update_layout(title=title)
        figures[gt] = fig
    return figures


def multi_channel_figure(
        sdt_metrics: pd.DataFrame,
        metric: str,
        title: str = "",
        yaxis_title: str = "",
        error_bars: Optional[str] = None,
        colors: u.COLORMAP_TYPE = None,
        show_other_gt: bool = True,
        show_err_bands: bool = False,
        subplots_vspace: float = 0.025,
        subplots_hspace: float = 0.01,
) -> go.Figure:
    subframe = _extract_sdt_subframe(sdt_metrics, metrics=metric, channel_type=None, threshold=None)
    subframe = subframe.droplevel(peyes.constants.METRIC_STR, axis=0)  # remove single-value levels from index
    gt_cols = subframe.columns.get_level_values(u.GT_STR).unique()
    assert 0 < len(gt_cols) <= 2, "Only 1 or 2 GT labelers are allowed for multi-channel figure"
    channel_types = subframe.index.get_level_values(peyes.constants.CHANNEL_TYPE_STR).unique()
    fig = make_subplots(
        rows=len(channel_types), cols=len(gt_cols), shared_xaxes="all", shared_yaxes="all",
        vertical_spacing=subplots_vspace, horizontal_spacing=subplots_hspace,
        row_titles=list(map(lambda ch: ch.title(), channel_types)),
        column_titles=list(map(lambda gt: gt.upper(), gt_cols)),
    )
    for r, ch_type in enumerate(channel_types):
        for c, gt in enumerate(gt_cols):
            data = subframe.xs(ch_type, level=peyes.constants.CHANNEL_TYPE_STR, axis=0).xs(gt, level=u.GT_STR, axis=1)
            detectors = u.sort_labelers(data.columns.get_level_values(u.PRED_STR).unique())
            for k, det in enumerate(detectors):
                if det in gt_cols:
                    if show_other_gt:
                        # current detector is a GT labeler, and we want to refer to it as "Other GT"
                        det_name = "Other GT"
                        det_color = "#bab0ac"
                        dash = "dot"
                    else:
                        # current detector is a GT labeler, and we don't want to show it in the figure
                        continue
                else:
                    # current detector is a prediction labeler (detection algorithm)
                    det_name = det.strip().removesuffix("Detector")
                    det_color = u.get_labeler_color(det_name, k, colors)
                    dash = None
                det_data = data.xs(det, level=u.PRED_STR, axis=1)
                thresholds = det_data.index.get_level_values(peyes.constants.THRESHOLD_STR).unique()
                mean = det_data.mean(axis=1)
                errors = h.calc_error_bars(det_data, error_bars)
                fig.add_trace(
                    row=r + 1, col=c + 1, trace=go.Scatter(
                        x=thresholds, y=mean, error_y=dict(type="data", array=errors),
                        name=det_name, legendgroup=det_name,
                        mode="lines+markers",
                        marker=dict(size=5, color=det_color),
                        line=dict(color=det_color, dash=dash),
                        showlegend=(c == 0 and r == 0),
                    )
                )
                if show_err_bands and errors is not None and not errors.isna().all():
                    y_upper, y_lower = mean + errors, mean - errors
                    fig.add_trace(
                        row=r + 1, col=c + 1, trace=go.Scatter(
                            x=np.concatenate((thresholds, thresholds[::-1])),
                            y=np.concatenate((y_upper, y_lower[::-1])),
                            fill="toself", fillcolor=det_color, opacity=0.2,
                            line=dict(color=det_color, width=0),
                            name=det_name, legendgroup=det_name, showlegend=False, hoverinfo="skip",
                        )
                    )
            if c == 0:
                yaxis_title = yaxis_title if yaxis_title else metric.replace("_", " ").lower()
                fig.update_yaxes(title_text=yaxis_title, row=r + 1, col=1)
            if r == len(channel_types) - 1:
                fig.update_xaxes(title_text="Threshold (samples)", row=len(channel_types), col=c + 1)
    fig.update_layout(
        title=title if title else f"Samples Channel :: {metric.replace("_", " ").title()} for Increasing Thresholds",
        legend=dict(orientation="h", yanchor="bottom", y=-0.3, xanchor="left", x=0.25),
    )
    return fig


def _extract_sdt_subframe(
        sdt_metrics: pd.DataFrame,
        channel_type: Optional[str] = None,
        threshold: Optional[Union[int, Sequence[int]]] = None,
        metrics: Optional[Union[str, Sequence[str]]] = None,
) -> pd.DataFrame:
    """
    Given a DataFrame with SDT metrics that has multi-level index (channel_type, metric, threshold), extract a sub-frame
    based on the provided parameters. Returns a DataFrame with the same index structure as the input DataFrame.

    :param sdt_metrics: pd.DataFrame; SDT metrics
    :param channel_type: str; channel type (onset/offset)
    :param threshold: int or list of int; threshold value(s)
    :param metrics: str or list of str; metric(s) to extract

    :return: pd.DataFrame; sub-frame with SDT metrics
    """
    sub_frame = sdt_metrics
    if channel_type:
        sub_frame = h.extract_subframe(
            sub_frame, level=peyes.constants.CHANNEL_TYPE_STR, value=channel_type, axis=0, drop_single_values=False
        )
    if threshold:
        sub_frame = h.extract_subframe(
            sub_frame, level=peyes.constants.THRESHOLD_STR, value=threshold, axis=0, drop_single_values=False
        )
    if metrics:
        sub_frame = h.extract_subframe(
            sub_frame, level=peyes.constants.METRIC_STR, value=metrics, axis=0, drop_single_values=False
        )
    return sub_frame



import numpy as np
import pandas as pd

from peyes._utils import constants as cnst
from peyes._utils.event_utils import calculate_num_samples
from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum, EventLabelSequenceType


def summarize_events(
        events: EventSequenceType,
) -> pd.DataFrame:
    """ Converts the given events to a DataFrame, where each row is an event and columns are event features. """
    if len(events) == 0:
        return pd.DataFrame()
    summaries = [e.summary() for e in events]
    return pd.DataFrame(summaries)


def events_to_labels(events: EventSequenceType, sampling_rate: float, min_num_samples=None) -> EventLabelSequenceType:
    """
    Converts the given events to a sequence of labels, where each event is mapped to a sequence of labels with length
    matching the number of samples in the event's duration (rounded up to the nearest integer).
    Samples with no event are labeled as `EventLabelEnum.UNDEFINED`.

    :param events: array-like of Event objects
    :param sampling_rate: the sampling rate of the output labels
    :param min_num_samples: the minimal number of samples in the output sequence. If None, the number of samples is
        determined by the total duration of the provided events.

    :return: sequence of labels
    """
    global_start_time = min(e.start_time for e in events)
    global_end_time = max(e.end_time for e in events)
    num_samples = calculate_num_samples(global_start_time, global_end_time, sampling_rate, min_num_samples)
    out = np.full(num_samples, EventLabelEnum.UNDEFINED)
    for e in events:
        corrected_start_time, corrected_end_time = e.start_time - global_start_time, e.end_time - global_start_time
        start_sample = int(np.round(corrected_start_time * sampling_rate / cnst.MILLISECONDS_PER_SECOND))
        end_sample = int(np.round(corrected_end_time * sampling_rate / cnst.MILLISECONDS_PER_SECOND))
        out[start_sample:end_sample] = e.label
    return out



import os
from typing import Tuple, Sequence

import cv2
import numpy as np
from tqdm import trange

import peyes._DataModels.config as cnfg
import peyes._utils.visualization_utils as vis_utils
from peyes._utils.event_utils import calculate_sampling_rate

_DEFAULT_EXTENSION = ".mp4"
_DEFAULT_CODEC = cv2.VideoWriter_fourcc(*"mp4v")


def create_video(
        t: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
        labels: np.ndarray,
        output_path: str,
        resolution: Tuple[int, int],
        bg_image: np.ndarray = None,
        bg_image_format: str = "BGR",
        label_colors: vis_utils.LabelColormapType = None,
        gaze_radius: int = 10,
        codec: int = _DEFAULT_CODEC,
        extension: str = _DEFAULT_EXTENSION,
        verbose: bool = False,
) -> str:
    """
    Creates a video file from gaze data.

    :param t: array of timestamps (ms)
    :param x: array of x-coordinates (pixels)
    :param y: array of y-coordinates (pixels)
    :param labels: array of event labels
    :param output_path: full path to the output video file
    :param resolution: tuple of (width, height) in pixels

    Optional Parameters:
    :param bg_image: background image (numpy array). If None, a black background will be used (default).
    :param bg_image_format: color format (RGB/BGR) for the background image, if provided. Default is BGR.
    :param label_colors: dictionary mapping event labels to hex/rgb colors. Default is the event color mapping from `config.py`.
    :param gaze_radius: radius of the gaze point in pixels. Default is 10.
    :param codec: codec for the video writer. Default is `mp4v`.
    :param extension: file extension for the output video file. Default is `.mp4`.
    :param verbose: if True, prints progress messages. Default is False.

    :return: full path to the output video file
    """
    assert len(t) == len(x) == len(y) == len(labels), "All input arrays must have the same length."
    fps = round(calculate_sampling_rate(t))
    frames = create_frames(x, y, labels, resolution, bg_image, bg_image_format, label_colors, gaze_radius, verbose)
    return _write_video(frames, output_path, fps, codec, extension, verbose)


def create_frames(
        x: np.ndarray,
        y: np.ndarray,
        labels: np.ndarray,
        resolution: Tuple[int, int],
        bg_image: np.ndarray = None,
        bg_image_format: str = "BGR",
        bg_image_alpha: float = 1,
        label_colors: vis_utils.LabelColormapType = None,
        gaze_radius: int = 10,
        verbose: bool = False,
) -> Sequence[np.ndarray]:
    """
    Creates a sequence of frames (numpy arrays) from gaze data.

    :param x: array of x-coordinates (pixels)
    :param y: array of y-coordinates (pixels)
    :param labels: array of event labels
    :param resolution: tuple of (width, height) in pixels

    Optional Parameters:
    :param bg_image: background image (numpy array). If None, a black background will be used (default).
    :param bg_image_format: color format (RGB/BGR) for the background image, if provided. Default is BGR.
    :param bg_image_alpha: alpha (opacity) value of the background image (range [0, 1]). Default is 1 (100% opacity).
    :param label_colors: dictionary mapping event labels to hex/rgb colors. If a label is missing, the default color is used.
    :param gaze_radius: radius of the gaze point in pixels. Default is 10.
    :param verbose: if True, prints progress messages. Default is False.

    :return: list of frames (numpy arrays)
    """
    assert len(x) == len(y) == len(labels), "All input arrays must have the same length."
    frames = []
    n_samples = len(x)
    bg_image = vis_utils.create_image(resolution, bg_image, bg_image_alpha, bg_image_format, "#000000")
    label_colors = vis_utils.get_label_colormap(label_colors)
    for i in trange(n_samples, desc="Creating Frames", disable=not verbose):
        curr_img = bg_image.copy()
        curr_x, curr_y = int(x[i]), int(y[i])
        color = label_colors.get(labels[i], label_colors[cnfg.EventLabelEnum.UNDEFINED])
        cv2.circle(curr_img, (curr_x, curr_y), gaze_radius, color, -1)
        frames.append(curr_img)
    return frames


def _write_video(
        frames: Sequence[np.ndarray],
        output_path: str,
        fps: int = 30,
        codec: int = _DEFAULT_CODEC,
        extension: str = _DEFAULT_EXTENSION,
        verbose: bool = False,
) -> str:
    if not output_path.endswith(extension):
        output_path += extension
    os.makedirs(os.path.dirname(output_path), exist_ok=True)
    if verbose:
        print(f"Writing video to: {output_path}")
    h, w, _ = frames[0].shape
    writer = cv2.VideoWriter(output_path, codec, fps, (w, h))
    for i in trange(len(frames), desc="Writing Frames", disable=not verbose):
        writer.write(frames[i])
    writer.release()
    if verbose:
        print("Video writing complete.")
    return output_path





import plotly.io as pio

import peyes
from analysis._default_values._helpers import *
import analysis.statistics.channel_sdt as ch_sdt

pio.renderers.default = "browser"
THRESHOLD = 5  # samples

# %%
#######################
## Fixation Channels ##

LABEL = 1       # EventLabelEnum.FIXATION.value

fixation_csdt_metrics = ch_sdt.load(
    dataset_name=DATASET_NAME,
    output_dir=PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=STIMULUS_TYPE,
    channel_type=None
)

## Stats

fix_onset_statistics, fix_onset_pvalues, fix_onset_dunns, fix_onset_Ns = ch_sdt.kruskal_wallis_dunns(
    fixation_csdt_metrics, "onset", THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

fix_offset_statistics, fix_offset_pvalues, fix_offset_dunns, fix_offset_Ns = ch_sdt.kruskal_wallis_dunns(
    fixation_csdt_metrics, "offset", THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

### Show Figures

fixation_dprime_figure = ch_sdt.multi_channel_figure(
    fixation_csdt_metrics,
    metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
fixation_dprime_figure.update_layout(width=1400, height=500,)
fixation_dprime_figure.show()

fixation_criterion_figure = ch_sdt.multi_channel_figure(
    fixation_csdt_metrics,
    metric=peyes.constants.CRITERION_STR,
    yaxis_title="Criterion", show_other_gt=True, show_err_bands=True
)
fixation_criterion_figure.update_layout(width=1400, height=500,)
fixation_criterion_figure.show()

# %%
######################
## Saccade Channels ##

LABEL = 2       # EventLabelEnum.FIXATION.value

saccade_csdt_metrics = ch_sdt.load(
    dataset_name=DATASET_NAME,
    output_dir=PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=STIMULUS_TYPE,
    channel_type=None
)

## Stats

sac_onset_statistics, sac_onset_pvalues, sac_onset_dunns, sac_onset_Ns = ch_sdt.kruskal_wallis_dunns(
    saccade_csdt_metrics, "onset", THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

sac_offset_statistics, sac_offset_pvalues, sac_offset_dunns, sac_offset_Ns = ch_sdt.kruskal_wallis_dunns(
    saccade_csdt_metrics, "offset", THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

### Show Figures

saccade_dprime_figure = ch_sdt.multi_channel_figure(
    saccade_csdt_metrics,
    metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
saccade_dprime_figure.update_layout(width=1400, height=500,)
saccade_dprime_figure.show()

saccade_criterion_figure = ch_sdt.multi_channel_figure(
    saccade_csdt_metrics,
    metric=peyes.constants.CRITERION_STR,
    yaxis_title="Criterion", show_other_gt=True, show_err_bands=True
)
saccade_criterion_figure.update_layout(width=1400, height=500,)
saccade_criterion_figure.show()

# %%
######################
## Save Figures

peyes.visualize.save_figure(
    fixation_dprime_figure, "alignment_metrics-fixation-dprime", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    fixation_criterion_figure, "alignment_metrics-fixation-criterion", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    saccade_dprime_figure, "alignment_metrics-saccade-dprime", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    saccade_criterion_figure, "alignment_metrics-saccade-criterion", FIGURES_DIR,
    as_png=True, as_html=False, as_json=False
)



#
# The following scripts are meant to compare velocity thresholds of the I-DT algorithm:
#   - Andersson et al. (2017) - fixation dispersion criteria: 2.7 deg + 55 ms; saccade velocity threshold: 45 deg/s
#   - Komoogortsev & Karpov (2013) - fixation dispersion criteria: 2 deg + 110 ms; saccade velocity threshold: 45 deg/s
#
# We evaluate detection performance compared to human coders on the lund2013+image dataset (only image stimuli).
# The following metrics are used:
#   1) overall sample-level agreement - balanced accuracy, Cohen's Kappa, MCC, 1-NLD
#   2) fixation onset/offset discriminability (d')
#   3) saccade onset/offset discriminability (d')
### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ###

import pandas as pd
import plotly.io as pio

import peyes

import analysis._parameter_exploration._helpers as h

pio.renderers.default = "browser"

COMPARED_DETECTORS = [det for det in h.DETECTORS.keys() if "idvt" in det.lower()]

# %%
# SAMPLE-LEVEL AGREEMENT (GLOBAL MEASURES)

import analysis.statistics.sample_metrics as sm

# Load Data
global_metrics = sm.load_global_metrics(
    h.DATASET_NAME, h.PROCESSED_DATA_DIR, stimulus_type=h.STIMULUS_TYPE, metric=None, iteration=1,
)
global_metrics = pd.concat(
    [global_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
).drop(index=peyes.constants.ACCURACY_STR, inplace=False)    # drop Accuracy metric

global_statistics, global_pvalues, global_dunns, global_Ns = sm.kruskal_wallis_dunns(
    global_metrics, [h.GT1, h.GT2], multi_comp=h.MULTI_COMP
)

global_distribution_figure = sm.global_metrics_distributions_figure(global_metrics, h.GT1, h.GT2,)
global_distribution_figure.show()

print("No significant differences in global metrics")

# %%
# ONSET & OFFSET DETECTION (FIXATIONS)

import analysis.statistics.channel_sdt as ch_sdt

LABEL = 1       # EventLabelEnum.FIXATION.value
THRESHOLD = 5   # samples

# Load Data
fixation_csdt_metrics = ch_sdt.load(
    dataset_name=h.DATASET_NAME, output_dir=h.PROCESSED_DATA_DIR, label=LABEL, stimulus_type=h.STIMULUS_TYPE,
)
fixation_csdt_metrics = pd.concat(
    [fixation_csdt_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
)

# calc stats (Kruskal-Wallis)
fix_onset_statistics, fix_onset_pvalues, fix_onset_dunns, fix_onset_Ns = ch_sdt.kruskal_wallis_dunns(
    fixation_csdt_metrics, "onset", THRESHOLD, [h.GT1, h.GT2], multi_comp=h.MULTI_COMP
)

fix_offset_statistics, fix_offset_pvalues, fix_offset_dunns, fix_offset_Ns = ch_sdt.kruskal_wallis_dunns(
    fixation_csdt_metrics, "offset", THRESHOLD, [h.GT1, h.GT2], multi_comp=h.MULTI_COMP
)

# Show Figures
fixation_dprime_figure = ch_sdt.multi_channel_figure(
    fixation_csdt_metrics, metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
fixation_dprime_figure.update_layout(width=1400, height=500,)
fixation_dprime_figure.show()

print("No significant differences in fixation onset/offset detection")

# %%
# ONSET & OFFSET DETECTION (SACCADES)

import analysis.statistics.channel_sdt as ch_sdt

LABEL = 2       # EventLabelEnum.SACCADE.value
THRESHOLD = 5   # samples

# Load Data
saccade_csdt_metrics = ch_sdt.load(
    dataset_name=h.DATASET_NAME, output_dir=h.PROCESSED_DATA_DIR, label=LABEL, stimulus_type=h.STIMULUS_TYPE,
)
saccade_csdt_metrics = pd.concat(
    [saccade_csdt_metrics.xs(det, axis=1, level='pred', drop_level=False) for det in COMPARED_DETECTORS], axis=1
)

# calc stats (Kruskal-Wallis)
sac_onset_statistics, sac_onset_pvalues, sac_onset_dunns, sac_onset_Ns = ch_sdt.kruskal_wallis_dunns(
    saccade_csdt_metrics, "onset", THRESHOLD, [h.GT1, h.GT2], multi_comp=h.MULTI_COMP
)

sac_offset_statistics, sac_offset_pvalues, sac_offset_dunns, sac_offset_Ns = ch_sdt.kruskal_wallis_dunns(
    saccade_csdt_metrics, "offset", THRESHOLD, [h.GT1, h.GT2], multi_comp=h.MULTI_COMP
)

# Show Figures
saccade_dprime_figure = ch_sdt.multi_channel_figure(
    saccade_csdt_metrics, metric=peyes.constants.D_PRIME_STR,
    yaxis_title=r"$d'$", show_other_gt=True, show_err_bands=True
)
saccade_dprime_figure.update_layout(width=1400, height=500,)
saccade_dprime_figure.show()

print(
    "saccade onset using I-DVT with Andersson params is *significantly* worse than with Komogorob & "
    "Komogorov-Birawo params (which are the same)"
)



import os
import warnings

import pandas as pd
import plotly.io as pio

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, STIMULUS_TYPE, PROCESSED_DATA_DIR, FIGURES_DIR, GT1

pio.renderers.default = "browser"

######################
# Load Events Data

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}

######################
# Create Figures for All Labelers

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    events_summary_figures = {
        lblr: peyes.visualize.event_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Events Summary",
        ) for lblr in all_labelers
    }
    fixations_summary_figures = {
        lblr: peyes.visualize.fixation_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Fixations Summary",
        ) for lblr in all_labelers
    }
    saccades_summary_figures = {
        lblr: peyes.visualize.saccade_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Saccades Summary",
        ).update_annotations(
            # move the "Azimuth" subtitle a bit up to avoid overlap with the 90° symbol
            y=0.4, selector={'text': peyes.constants.AZIMUTH_STR.title()}
        ) for lblr in all_labelers
    }

######################
# Show Figures Single Labeler

for lblr in all_labelers:
    if lblr != GT1:
        # avoid showing all labelers' figures to avoid cluttering the browser
        # TODO: remove this condition if you want to see all labelers' figures
        continue
    events_summary_figures[lblr].show()
    fixations_summary_figures[lblr].show()
    saccades_summary_figures[lblr].show()

######################
# Save Figures Single Labeler

for lblr in all_labelers:
    if lblr != GT1:
        # avoid saving all labelers' figures to avoid cluttering the filesystem
        # TODO: remove this condition if you want to save all labelers' figures
        continue
    peyes.visualize.save_figure(
        events_summary_figures[lblr],
        f"events_summary_{lblr}",
        FIGURES_DIR,
        as_png=True, as_html=False, as_json=False
    )
    peyes.visualize.save_figure(
        fixations_summary_figures[lblr],
        f"fixations_summary_{lblr}",
        FIGURES_DIR,
        as_png=True, as_html=False, as_json=False
    )
    peyes.visualize.save_figure(
        saccades_summary_figures[lblr],
        f"saccades_summary_{lblr}",
        FIGURES_DIR,
        as_png=True, as_html=False, as_json=False
    )



from typing import Union, Sequence

import numpy as np

from peyes._DataModels.Event import BaseEvent
from peyes._DataModels.EventLabelEnum import EventLabelEnum

UnparsedEventLabelType = Union[EventLabelEnum, BaseEvent, int, str, float, np.number]
UnparsedEventLabelSequenceType = Sequence[UnparsedEventLabelType]



import os

import numpy as np
import pandas as pd
import cv2
import plotly.io as pio

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, STIMULUS_TYPE, PROCESSED_DATA_DIR, FIGURES_DIR

pio.renderers.default = "browser"

STIMULUS_DIR = os.path.join(u.BASE_DIR, "stimuli", DATASET_NAME.capitalize(), STIMULUS_TYPE)

######################
# Load Raw Data

dataset = u.load_dataset(DATASET_NAME, verbose=False)
img_trial_id = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)[5]
trial_data = dataset[dataset[peyes.constants.TRIAL_ID_STR] == img_trial_id]

img_name = trial_data[peyes.constants.STIMULUS_NAME_STR].values[0]
img_name += ".png"
img = cv2.imread(os.path.join(STIMULUS_DIR, img_name))
resolution = (img.shape[1], img.shape[0])

pixel_size = trial_data[peyes.constants.PIXEL_SIZE_STR].values[0]
viewer_distance = trial_data[peyes.constants.VIEWER_DISTANCE_STR].values[0]
t = trial_data[peyes.constants.T].values
x = trial_data[peyes.constants.X].values
y = trial_data[peyes.constants.Y].values

x[(x < 0) | (x >= resolution[0])] = np.nan
y[(y < 0) | (y >= resolution[1])] = np.nan

######################
# Load Labels

labels_df = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, peyes.constants.LABELS_STR + ".pkl"))
labels_df = labels_df.xs(1, level=peyes.constants.ITERATION_STR, axis=1)  # Only use first iteration
labels_df = labels_df.xs(img_trial_id, level=peyes.constants.TRIAL_ID_STR, axis=1)    # Only use the relevant trial
labels_df = labels_df.dropna(how="all", axis=0)   # Drop rows with all NaNs

labeler_names = u.sort_labelers(labels_df.columns.get_level_values(peyes.constants.LABELER_STR).unique())

######################

top_fig = peyes.visualize.gaze_trajectory(
    x=x, y=y, resolution=resolution, title="Gaze Trajectory",
    bg_image=img, bg_image_format='rgb', bg_alpha=0.5, marker_alpha=1,
    t=t, colorscale='Jet'
)
top_fig.update_layout(
    title=None,
)
top_fig.show()


middle_fig = peyes.visualize.gaze_over_time(
    x=x, y=y, t=t, resolution=resolution, title="Gaze Over Time",
    v=peyes._utils.pixel_utils.calculate_velocities(x, y, t), v_measure='px/s'
)
middle_fig.update_layout(
    paper_bgcolor='rgba(0, 0, 0, 0)',
    plot_bgcolor='rgba(0, 0, 0, 0)',
    height=top_fig.layout.height // 2,
    width=top_fig.layout.width,
    title=None,
)
middle_fig.show()

bottom_fig = peyes.visualize.scarfplot_comparison_figure(
    t,
    *[labels_df[labeler_name] for labeler_name in labeler_names],
    names=labeler_names,
)
bottom_fig.update_layout(
    height=top_fig.layout.height // 2,
    width=top_fig.layout.width,
    title=None,
)
bottom_fig.show()

######################
# Save Figures

peyes.visualize.save_figure(
    top_fig, "single_trial-gaze_trajectory", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    middle_fig, "single_trial-gaze_over_time", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)
peyes.visualize.save_figure(
    bottom_fig, "single_trial-scarfplot", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)



from enum import IntEnum
from typing import Sequence


class EventLabelEnum(IntEnum):
    UNDEFINED = 0
    FIXATION = 1
    SACCADE = 2
    PSO = 3
    SMOOTH_PURSUIT = 4
    BLINK = 5


EventLabelSequenceType = Sequence[EventLabelEnum]



from typing import Optional, Union, Tuple, Sequence

import pandas as pd
import scipy.stats as stats
import statsmodels.stats.multitest as multi
import plotly.graph_objects as go

import peyes
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType

import analysis.utils as u
import analysis.statistics._helpers as h

########################


def load(
        dataset_name: str,
        output_dir: str,
        label: Optional[Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType]] = None,
        stimulus_type: Optional[Union[str, Sequence[str]]] = None,
        matching_schemes: Optional[Union[str, Sequence[str]]] = None,
        features: Optional[Union[str, Sequence[str]]] = None,
) -> pd.DataFrame:
    if matching_schemes is None or len(matching_schemes) == 0:
        matching_schemes = None
    elif isinstance(matching_schemes, str):
        matching_schemes = [matching_schemes]
    matched_features = h.load_data(
        dataset_name=dataset_name, output_dir=output_dir,
        data_dir_name=f"{peyes.constants.MATCHES_STR}_{peyes.constants.METRICS_STR}",
        label=label, stimulus_type=stimulus_type, sub_index=matching_schemes,
        filename_suffix=f"matched_{peyes.constants.FEATURES_STR}", iteration=1,
    )
    if features is None or len(features) == 0:
        return matched_features
    if isinstance(features, str):
        features = [features]
    is_features = matched_features.index.get_level_values(peyes.constants.FEATURE_STR).isin(features)
    return matched_features.loc[is_features]


def kruskal_wallis_dunns(
        matched_features: pd.DataFrame,
        matching_scheme: str,
        gt_cols: Union[str, Sequence[str]],
        features: Union[str, Sequence[str]] = None,
        multi_comp: Optional[str] = "fdr_bh",
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    For each unique value in the input DataFrame's index (feature names) and each of the GT labelers, performs
    Kruskal-Wallis test with post-hoc Dunn's test for multiple comparisons.
    Returns the KW-statistic, KW-p-value, Dunn's-p-values and number of samples for each (index, GT labeler) pair.
    """
    subframe = h.extract_subframe(
        matched_features, level=u.MATCHING_SCHEME_STR, value=matching_scheme, axis=0, drop_single_values=True
    )
    if features:
        subframe = h.extract_subframe(
            subframe, level=peyes.constants.FEATURE_STR, value=features, axis=0, drop_single_values=False
        )
    statistics, pvalues, dunns, Ns = h.kruskal_wallis_dunns(subframe, gt_cols, multi_comp)
    return statistics, pvalues, dunns, Ns


def wilcoxon_signed_rank(
        matched_features: pd.DataFrame,
        matching_scheme: str,
        gt_col: str,
        features: Union[str, Sequence[str]] = None,
        multi_comp: str = "fdr_bh",
        alpha: float = 0.05,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    assert 0 < alpha < 1, "Argument `alpha` must be in the open interval (0, 1)"
    scheme_matched_features = h.extract_subframe(
        matched_features, level=u.MATCHING_SCHEME_STR, value=matching_scheme, axis=0, drop_single_values=True
    )
    if features:
        scheme_matched_features = h.extract_subframe(
            scheme_matched_features, level=peyes.constants.FEATURE_STR, value=features, axis=0, drop_single_values=False
        )
    features = sorted(
        scheme_matched_features.index.unique(),
        key=lambda met: u.METRICS_CONFIG[met][1] if met in u.METRICS_CONFIG else ord(met[0])
    )
    gt_columns = scheme_matched_features.columns.get_level_values(u.GT_STR).unique()
    statistics, pvalues, Ns = {}, {}, {}
    for i, feat in enumerate(features):
        alternative = 'two-sided' if peyes.constants.DIFFERENCE_STR in feat else 'greater'
        gt_series = scheme_matched_features.xs(gt_col, level=u.GT_STR, axis=1).loc[feat]
        gt_df = gt_series.unstack().drop(columns=gt_columns, errors='ignore')
        for j, det in enumerate(gt_df.columns):
            vals = gt_df[det].explode().dropna().values.astype(float)
            if feat in [f"{peyes.constants.TIME_STR}_overlap", f"{peyes.constants.TIME_STR}_iou"]:
                # for IoU & (normalized) Overlap, the best values are 1, and Wilcoxon's needs to compare against 0
                vals = 1 - vals
            statistic, pvalue = stats.wilcoxon(x=vals, alternative=alternative, zero_method='zsplit')
            statistics[(feat, det)] = statistic
            pvalues[(feat, det)] = pvalue
            Ns[(feat, det)] = len(vals)
    statistics = pd.Series(statistics).unstack()
    pvalues = pd.Series(pvalues).unstack()
    Ns = pd.Series(Ns).unstack()
    corrected_pvalues = pd.DataFrame(
        index=pvalues.index,
        data={col: multi.multipletests(pvalues[col], method=multi_comp, alpha=alpha)[1] for col in pvalues.columns}
    )
    return statistics, pvalues, corrected_pvalues, Ns
# TODO: add similar function comparing (GT1, Pred) against (GT1, GT2) for each trial/detector


def distributions_figure(
        matched_features: pd.DataFrame,
        matching_scheme: str,
        gt1: str,
        gt2: str,
        features: Union[str, Sequence[str]] = None,
        colors: u.COLORMAP_TYPE = None,
        title: Optional[str] = None,
        show_other_gt: bool = False,
        share_x: bool = False,
        share_y: bool = False,
) -> go.Figure:
    subframe = h.extract_subframe(
        matched_features, level=u.MATCHING_SCHEME_STR, value=matching_scheme, axis=0, drop_single_values=True
    )
    if features:
        subframe = h.extract_subframe(
            subframe, level=peyes.constants.FEATURE_STR, value=features, axis=0, drop_single_values=False
        )
    if not title:
        split_scheme = matching_scheme.split("_")
        scheme_name, scheme_threshold = "_".join(split_scheme[:-1]), int(split_scheme[-1])
        threshold_str = f"{scheme_threshold:.2f}(a.u.)" if isinstance(scheme_threshold, float) else f"{scheme_threshold} (samples)"
        title = ("Matched Events :: Features <br>" +
                 f"<sup>({u.MATCHING_SCHEME_STR.replace("_", " ").title()}: {scheme_name.replace("_", " ").title()}  " +
                 f"Threshold: {threshold_str}</sup>")
    fig = h.distributions_figure(
        subframe, gt1=gt1, gt2=gt2, colors=colors, title=title,
        show_other_gt=show_other_gt, share_x=share_x, share_y=share_y,
    )
    return fig



import os

import pandas as pd
import plotly.io as pio

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, STIMULUS_TYPE, PROCESSED_DATA_DIR, FIGURES_DIR, GT1

from peyes._DataModels.EventLabelEnum import EventLabelEnum

pio.renderers.default = "browser"

######################
# Load Events Data

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)  # Keep only the first iteration
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}

######################
# Main Sequence for each Labeler

main_sequence_figures = {
    lblr: peyes.visualize.main_sequence(
        saccades=events_by_labelers[lblr][events_by_labelers[lblr].apply(lambda e: e.label == EventLabelEnum.SACCADE)],
        include_outliers=False,
    ) for lblr in all_labelers
}

######################
# Show Stat Results & Figure for Single Labeler

main_sequence_figures[GT1][0].show()
print(main_sequence_figures[GT1][1].iloc[0, 0].summary())    # print the statistical results for the inliers' m.s.

######################
# Save Figures

peyes.visualize.save_figure(
    main_sequence_figures[GT1][0], "main_sequence", FIGURES_DIR, as_png=True, as_html=False, as_json=False
)



import os
import warnings

import pandas as pd
import plotly.io as pio
from docutils.nodes import legend

import peyes
import analysis.utils as u
from analysis._default_values._helpers import DATASET_NAME, PROCESSED_DATA_DIR, ALPHA

pio.renderers.default = "browser"

######################

LABEL = 1
STIMULUS_TYPE = peyes.constants.IMAGE_STR
GT1, GT2 = "RA", "MN"
MULTI_COMP = "fdr_bh"

# %%
##########################
##  Feature Comparison  ##

stim_trial_ids = u.get_trials_for_stimulus_type(DATASET_NAME, STIMULUS_TYPE)

all_events = pd.read_pickle(os.path.join(PROCESSED_DATA_DIR, DATASET_NAME, "events.pkl"))
all_events = all_events.xs(1, level=peyes.constants.ITERATION_STR, axis=1)
all_events = all_events.loc[:, all_events.columns.get_level_values(peyes.constants.TRIAL_ID_STR).isin(stim_trial_ids)]
all_events = all_events.dropna(axis=0, how="all")

all_labelers = all_events.columns.get_level_values(peyes.constants.LABELER_STR).unique()
events_by_labelers = {
    lblr: all_events.xs(lblr, level=peyes.constants.LABELER_STR, axis=1).stack().dropna() for lblr in all_labelers
}

fixations_comparison_figure = peyes.visualize.feature_comparison(
    [
        peyes.constants.DURATION_STR,
        peyes.constants.AMPLITUDE_STR,
        peyes.constants.PEAK_VELOCITY_STR,
        peyes.constants.MEDIAN_VELOCITY_STR,
        peyes.constants.COUNT_STR,
     ],
    *[vals[vals.apply(lambda e: e.label == 1)] for vals in events_by_labelers.values()],
    labels=events_by_labelers.keys(),
    title="Fixation Features Comparison",
    include_outliers=False,
)
fixations_comparison_figure.show()

saccades_comparison_figure = peyes.visualize.feature_comparison(
    [
        peyes.constants.DURATION_STR,
        peyes.constants.AMPLITUDE_STR,
        peyes.constants.PEAK_VELOCITY_STR,
        peyes.constants.MEDIAN_VELOCITY_STR,
        peyes.constants.COUNT_STR,
     ],
    *[vals[vals.apply(lambda e: e.label == 2)] for vals in events_by_labelers.values()],
    labels=events_by_labelers.keys(),
    title="Saccade Features Comparison",
    include_outliers=False,
)
saccades_comparison_figure.show()

# %%
##########################
##  Events Summary Plot  ##

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    events_summary_figures = {
        lblr: peyes.visualize.event_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Events Summary",
        ) for lblr in all_labelers
    }
    fixations_summary_figures = {
        lblr: peyes.visualize.fixation_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Fixations Summary",
        ) for lblr in all_labelers
    }
    saccades_summary_figures = {
        lblr: peyes.visualize.saccade_summary(
            events_by_labelers[lblr],
            show_outliers=True,
            title=f"{lblr} :: Saccades Summary",
        ) for lblr in all_labelers
    }

events_summary_figures[GT1].show()
fixations_summary_figures[GT1].show()
saccades_summary_figures[GT1].show()


# %%
######################
##  Sample Metrics  ##

import analysis.statistics.sample_metrics as sm

sample_global_metrics = sm.load_global_metrics(
    DATASET_NAME, PROCESSED_DATA_DIR, stimulus_type=STIMULUS_TYPE, metric=None
)
sm_global_statistics, sm_global_pvalues, sm_global_dunns, sm_global_Ns = sm.kruskal_wallis_dunns(
    sample_global_metrics, [GT1, GT2], multi_comp=MULTI_COMP
)
sm_global_metrics_fig = sm.global_metrics_distributions_figure(sample_global_metrics, GT1, gt2=GT2, only_box=False)
sm_global_metrics_fig.show()

###

sample_sdt_metrics = sm.load_sdt(
    DATASET_NAME, PROCESSED_DATA_DIR, label=LABEL, stimulus_type=STIMULUS_TYPE, metric=None
)
sm_sdt_statistics, sm_sdt_pvalues, sm_sdt_dunns, sm_sdt_Ns = sm.kruskal_wallis_dunns(
    sample_sdt_metrics, [GT1, GT2], multi_comp=MULTI_COMP
)
sample_sdt_metrics_fig = sm.sdt_distributions_figure(sample_sdt_metrics, GT1, gt2=GT2, only_box=False)
sample_sdt_metrics_fig.show()

# %%
##########################
##  Channel Time Diffs  ##

import analysis.statistics.channel_time_diffs as ctd

time_diffs = ctd.load(
    DATASET_NAME, PROCESSED_DATA_DIR, label=LABEL, stimulus_type=STIMULUS_TYPE
)
ctd_statistics, ctd_pvalues, ctd_dunns, ctd_Ns = ctd.kruskal_wallis_dunns(
    time_diffs, [GT1, GT2], multi_comp=MULTI_COMP
)
time_diffs_fig = ctd.distributions_figure(time_diffs, GT1, gt2=GT2, only_box=False)
time_diffs_fig.show()

# %%
###########################
##  Channel SDT Metrics  ##

import analysis.statistics.channel_sdt as csdt

THRESHOLD = 10  # samples

label_csdt_metrics = csdt.load(
    dataset_name=DATASET_NAME,
    output_dir=PROCESSED_DATA_DIR,
    label=LABEL,
    stimulus_type=STIMULUS_TYPE,
    channel_type=None
)

###
CHANNEL_TYPE = "onset"

csdt_onset_statistics, csdt_onset_pvalues, csdt_onset_dunns, csdt_onset_Ns = csdt.kruskal_wallis_dunns(
    label_csdt_metrics, CHANNEL_TYPE, THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

threshold_onset_fig = csdt.single_threshold_figure(label_csdt_metrics, CHANNEL_TYPE, THRESHOLD, GT1, gt2=GT2)
threshold_onset_fig.show()

csdt_onset_figs = csdt.multi_threshold_figures(
    label_csdt_metrics, CHANNEL_TYPE, show_other_gt=True, show_err_bands=True
)
csdt_onset_figs[GT1].show()

###
CHANNEL_TYPE = "offset"

csdt_offset_statistics, csdt_offset_pvalues, csdt_offset_dunns, csdt_offset_Ns = csdt.kruskal_wallis_dunns(
    label_csdt_metrics, CHANNEL_TYPE, THRESHOLD, [GT1, GT2], multi_comp=MULTI_COMP
)

threshold_offset_fig = csdt.single_threshold_figure(label_csdt_metrics, CHANNEL_TYPE, THRESHOLD, GT1, gt2=GT2)
threshold_offset_fig.show()

csdt_offset_figs = csdt.multi_threshold_figures(label_csdt_metrics, CHANNEL_TYPE, show_err_bands=True)
csdt_offset_figs[GT1].show()

###
# Both GT, Both Channels, Only d'

csdt_multi_channel_fig = csdt.multi_channel_figure(
    label_csdt_metrics, peyes.constants.D_PRIME_STR, yaxis_title=r"$d'$", show_err_bands=True
)
csdt_multi_channel_fig.show()


# %%
###############################
##  Matched-Events Features  ##

import analysis.statistics.matched_features as mf

SCHEME = "window_10"

matched_features = mf.load(
    dataset_name=DATASET_NAME,
    output_dir=PROCESSED_DATA_DIR,
    label=None,
    stimulus_type=STIMULUS_TYPE,
    matching_schemes=None,
)

mf_statistics, mf_pvalues, mf_dunns, mf_Ns = mf.kruskal_wallis_dunns(
    matched_features=matched_features, matching_scheme=SCHEME, gt_cols=[GT1, GT2], features=None, multi_comp=MULTI_COMP
)
mf_statistics2, mf_pvalues2, mf_corrected_pvalues2, mf_Ns2 = mf.wilcoxon_signed_rank(
    matched_features=matched_features, matching_scheme=SCHEME, gt_col=GT1, features=None, multi_comp=MULTI_COMP, alpha=ALPHA
)

mf_fig = mf.distributions_figure(matched_features, SCHEME, GT1, GT2)
mf_fig.show()


# %%
########################
## Matched-Events SDT ##

import analysis.statistics.matched_sdt as msdt

SCHEME = "window_10"

matched_sdt = msdt.load(
    dataset_name=DATASET_NAME, output_dir=PROCESSED_DATA_DIR,
    label=LABEL, stimulus_type=STIMULUS_TYPE, matching_schemes=None, metrics=None
)

msdt_statistics, msdt_pvalues, msdt_dunns, msdt_Ns = msdt.kruskal_wallis_dunns(
    matched_sdt, SCHEME, [GT1, GT2], metrics=None, multi_comp=MULTI_COMP
)

msdt_single_fig = msdt.single_scheme_figure(
    matched_sdt, SCHEME, GT1, gt2=GT2, metrics=None, only_box=False
)
msdt_single_fig.show()

msdt_figs = msdt.multi_threshold_figures(
    matched_sdt, "window", metrics=None, show_other_gt=True, show_err_bands=True
)
msdt_figs[GT1].show()

###
# Both GT, Both Channels, Only match-rate, f1, d'

msdt_multi_metric_fic = msdt.multi_metric_figure(
    matched_sdt, "window",
    [peyes.constants.MATCH_RATIO_STR, peyes.constants.F1_STR, peyes.constants.D_PRIME_STR],
    show_other_gt=True, show_err_bands=True
)
msdt_multi_metric_fic.show()



from peyes._utils.visualization_utils import save_figure
from peyes.visualize._gaze import gaze_trajectory, gaze_heatmap, gaze_over_time
from peyes.visualize._event_summary import event_summary, fixation_summary, saccade_summary
from peyes.visualize._features import feature_relationship, main_sequence, feature_comparison
from peyes.visualize._scarfplot import scarfplot_comparison_figure, add_scarfplot_to_figure
from peyes.visualize._video import create_video, create_frames



import peyes._utils.constants as cnst
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._utils.pixel_utils import calculate_pixel_size as _calc_ps

VIEWER_DISTANCE = 60    # cm

SCREEN_MONITOR = {
    cnst.WIDTH_STR: cnst.TOBII_WIDTH,
    cnst.HEIGHT_STR: cnst.TOBII_HEIGHT,
    cnst.RESOLUTION_STR: cnst.TOBII_RESOLUTION,
    cnst.PIXEL_SIZE_STR: _calc_ps(cnst.TOBII_WIDTH, cnst.TOBII_HEIGHT, cnst.TOBII_RESOLUTION),
}

EVENT_MAPPING = {
    EventLabelEnum.UNDEFINED: {
        cnst.LABEL_STR: EventLabelEnum.UNDEFINED.name,
        cnst.COLOR_STR: "#dddddd",
        cnst.MIN_DURATION_STR: 0,   # ms
        cnst.MAX_DURATION_STR: 1e9,  # ms
    },
    EventLabelEnum.FIXATION: {
        cnst.LABEL_STR: EventLabelEnum.FIXATION.name,
        cnst.COLOR_STR: "#1f78b4",
        cnst.MIN_DURATION_STR: 55,
        cnst.MAX_DURATION_STR: 2500,
    },
    EventLabelEnum.SACCADE: {
        cnst.LABEL_STR: EventLabelEnum.SACCADE.name,
        cnst.COLOR_STR: "#33a02c",
        cnst.MIN_DURATION_STR: 10,
        cnst.MAX_DURATION_STR: 200,
    },
    EventLabelEnum.PSO: {
        cnst.LABEL_STR: EventLabelEnum.PSO.name,
        cnst.COLOR_STR: "#b2df8a",
        cnst.MIN_DURATION_STR: 4,
        cnst.MAX_DURATION_STR: 40,
    },
    EventLabelEnum.SMOOTH_PURSUIT: {
        cnst.LABEL_STR: EventLabelEnum.SMOOTH_PURSUIT.name,
        cnst.COLOR_STR: "#fb9a99",
        cnst.MIN_DURATION_STR: 40,
        cnst.MAX_DURATION_STR: 5000,
    },
    EventLabelEnum.BLINK: {
        cnst.LABEL_STR: EventLabelEnum.BLINK.name,
        cnst.COLOR_STR: "#222222",
        cnst.MIN_DURATION_STR: 20,
        cnst.MAX_DURATION_STR: 2500,
    }
}



from typing import Dict, Union

import numpy as np
import pandas as pd
from tqdm import tqdm

import peyes._utils.constants as cnst
from peyes._DataModels.Event import EventSequenceType, EventLabelEnum


def features_by_labels(events: EventSequenceType) -> pd.DataFrame:
    """
    Aggregates the given events into a DataFrame, where each row is an event-label and columns are event features.
    Values of the same event-label are grouped together as lists (e.g. aggregated.loc["SACCADE", "amplitude"] is a list
    of all saccade amplitudes in the given events).
    """
    summary = pd.DataFrame([e.summary() for e in events])
    try:
        aggregated = summary.groupby(cnst.LABEL_STR).agg(list)    # rows are event labels, columns are features
    except KeyError:
        aggregated = pd.DataFrame(index=[l for l in EventLabelEnum], columns=[])
    for l in EventLabelEnum:
        if l not in aggregated.index:
            aggregated.loc[l] = [[] for _ in range(len(aggregated.columns))]
    # remove undefined events, these shouldn't exist in the first place:
    aggregated.drop(index=EventLabelEnum.UNDEFINED, inplace=True, errors="ignore")
    if aggregated.empty:
        return aggregated.sort_index()
    aggregated[cnst.COUNT_STR] = aggregated[cnst.DURATION_STR].map(len)
    return aggregated.sort_index()


def get_features(
        events: EventSequenceType, *features: str, verbose: bool = False
) -> Union[np.ndarray, Dict[str, np.ndarray]]:
    """
    Extracts the specified features from the given events.
    :param events: the events to extract features from
    :param features: the features to extract. Supported features are:
        - 'start_time' or 'onset': the start time of the event (ms)
        - 'end_time' or 'offset': the end time of the event (ms)
        - 'duration': the duration of the event (ms)
        - 'amplitude': the amplitude of the event (deg)
        - 'azimuth': the azimuth of the event (deg)
        - 'center_pixel' or 'center': the center pixel of the event
    :param verbose: if True, display a progress bar while extracting the features
    :return: the extracted features as a numpy array (for a single feature) or a dictionary of numpy arrays (for
        multiple features).
    """
    results: Dict[str, np.ndarray] = {}
    for feat in tqdm(features, desc="Event Features", disable=not verbose):
        results[feat] = _get_features_impl(events, feat)
    if len(results) == 1:
        return next(iter(results.values()))
    return results


def _get_features_impl(events: EventSequenceType, feature: str) -> np.ndarray:
    feature_lower = feature.lower().strip().replace(" ", "_").replace("-", "_")
    feature_lower = feature_lower.removesuffix('s')
    if feature_lower == "start_time" or feature_lower == "onset":
        return np.array([event.start_time for event in events])
    if feature_lower == "end_time" or feature_lower == "offset":
        return np.array([event.end_time for event in events])
    if feature_lower == "duration":
        return np.array([event.duration for event in events])
    if feature_lower == "amplitude":
        return np.array([event.amplitude for event in events])
    if feature_lower == "azimuth":
        return np.array([event.azimuth for event in events])
    if feature_lower == "center_pixel" or feature_lower == "center":
        return np.array([event.center_pixel for event in events])
    raise ValueError(f"Unknown event feature '{feature}'")



from typing import Union, Sequence

import numpy as np
import pandas as pd

import peyes._utils.constants as cnst
from peyes._DataModels.Event import EventSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelSequenceType

from peyes._base.create import create_boolean_channel
from peyes.alignment_metrics._timing_differences import timing_differences
from peyes._utils.metric_utils import dprime_and_criterion


def onset_detection_metrics(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        threshold: Union[int, Sequence[int]],
        sampling_rate: float = None,
        min_num_samples=None,
        dprime_correction: str = "loglinear",
) -> pd.DataFrame:
    """
    1. Converts the label/event sequences to a boolean array indicating onsets (MNE-style event channel).
    2. Matches between `True` indices (event onsets) of the two channels such that the difference between their indices
        is minimal.
    3. For each of threshold value, calculates:
        a) Contingency measures of the pairing (P, PP, TP, N)
        b) Signal Detection Theory (SDT) metrics (recall, precision, F1-score, false alarm rate, d-prime, criterion)

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param threshold: int or array-like of int; threshold values to determine if a pairing is valid
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects
    :param dprime_correction: str; optional correction method for floor/ceiling effects on the hit-rate and/or false-alarm rate
    :return: DataFrame of contingency measures and SDT metrics
    """
    return _signal_detection_metrics(
        ground_truth, prediction, threshold, "onset", sampling_rate, min_num_samples, dprime_correction
    )


def offset_detection_metrics(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        threshold: Union[int, Sequence[int]],
        sampling_rate: float = None,
        min_num_samples=None,
        dprime_correction: str = "loglinear",
) -> pd.DataFrame:
    """
    1. Converts the label/event sequences to a boolean array indicating offsets (MNE-style event channel).
    2. Matches between `True` indices (event offsets) of the two channels such that the difference between their indices
        is minimal.
    3. For each of threshold value, calculates:
        a) Contingency measures of the pairing (P, PP, TP, N)
        b) Signal Detection Theory (SDT) metrics (recall, precision, F1-score, false alarm rate, d-prime, criterion)

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param threshold: int or array-like of int; threshold values to determine if a pairing is valid
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects
    :param dprime_correction: str; optional correction method for floor/ceiling effects on the hit-rate and/or false-alarm rate
    :return: DataFrame of contingency measures and SDT metrics
    """
    return _signal_detection_metrics(
        ground_truth, prediction, threshold, "offset", sampling_rate, min_num_samples, dprime_correction
    )


def _signal_detection_metrics(
        ground_truth: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        prediction: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        threshold: Union[int, Sequence[int]],
        channel_type: str,
        sampling_rate: float = None,
        min_num_samples=None,
        dprime_correction: str = "loglinear",
) -> pd.DataFrame:
    """
    Converts the labels/events to boolean channels and matches between `True` indices of the two channels such that the
    difference between their indices is minimal. Then, for each threshold value, calculates the contingency measures
    of the pairing (P, PP, TP, N) and Signal Detection Theory (SDT) metrics (recall, precision, F1-score, false alarm
    rate, d-prime, criterion).
    Returns a DataFrame where the rows are the threshold values and the columns are the metrics.

    Note: False-Alarm rate could exceed 1.0 for large window-sizes, because PP events that "share" a window are
    counted separately, but null-windows (N) are counted once. This results in ((PP-TP)/N) exceeding the range [0, 1].
    This mostly happens when considering multiple positive labels, because their inter-event-interval may be smaller
    than the window-size.
    # TODO: Consider adding a "correction" parameter to handle this case.

    :param ground_truth: array-like of ground-truth labels or Event objects
    :param prediction: array-like of predicted labels or Event objects
    :param threshold: int or array-like of int; threshold values to determine if a pairing is valid
    :param channel_type: either 'onset' or 'offset'
    :param sampling_rate: sampling rate of the channels; only used if `ground_truth` or `prediction` are Event objects
    :param min_num_samples: if not None, marks the minimal the number of samples in the channels; only used if
        `ground_truth` or `prediction` are Event objects
    :param dprime_correction: str; optional correction method for floor/ceiling effects on the hit-rate and/or false-alarm rate

    :return: DataFrame of contingency measures and SDT metrics
    """
    gt_channel = create_boolean_channel(
        data=ground_truth, channel_type=channel_type, sampling_rate=sampling_rate, min_num_samples=min_num_samples
    )
    pred_channel = create_boolean_channel(
        data=prediction, channel_type=channel_type, sampling_rate=sampling_rate, min_num_samples=min_num_samples
    )
    p, pp = gt_channel.sum(), pred_channel.sum()  # number of positive samples in GT and prediction
    all_matched_diffs = timing_differences(
        ground_truth, prediction,
        max_diff=max(threshold) + 1,
        channel_type=channel_type,
        sampling_rate=sampling_rate,
        min_num_samples=min_num_samples
    )
    if isinstance(threshold, int):
        threshold = [threshold]
    results = {}
    for thresh in threshold:
        thresh_results = {"P": p, "PP": pp}
        tp = np.sum(np.abs(all_matched_diffs) <= thresh)
        thresh_results["TP"] = tp
        double_thresh = 2 * thresh + 1  # threshold can be before or after a particular sample
        n = max(0, (len(gt_channel) - double_thresh * p) / double_thresh)  # number of negative "windows" in GT channel
        thresh_results["N"] = n

        # true positive rate (TPR), sensitivity, hit-rate
        if p > 0 and 0 <= tp / p <= 1:
            recall = tp / p
        else:
            recall = np.nan
        # positive predictive value (PPV)
        if pp > 0 and 0 <= tp / pp <= 1:
            precision = tp / pp
        else:
            precision = np.nan
        # F1-score
        if np.isfinite(precision + recall) and ((precision + recall) > 0):
            f1_score = 2 * (precision * recall) / (precision + recall)
        else:
            f1_score = np.nan
        # FPR, type I error, 1 - specificity
        if n > 0 and 0 <= (pp - tp) / n <= 1:
            false_alarm_rate = (pp - tp) / n
        else:
            # FA rate exceed 1.0 due to over-counting false-alarm windows
            false_alarm_rate = np.nan
        d_prime, criterion = dprime_and_criterion(p, n, pp, tp, correction=dprime_correction)

        # update values:
        thresh_results[cnst.RECALL_STR] = recall
        thresh_results[cnst.PRECISION_STR] = precision
        thresh_results[cnst.F1_STR] = f1_score
        thresh_results[cnst.FALSE_ALARM_RATE_STR] = false_alarm_rate
        thresh_results[cnst.D_PRIME_STR] = d_prime
        thresh_results[cnst.CRITERION_STR] = criterion
        results[thresh] = thresh_results
    results = pd.DataFrame(results).T
    results.index.name = cnst.THRESHOLD_STR
    results.columns.name = cnst.METRIC_STR
    return results





from typing import Union

import numpy as np

import peyes._utils.constants as cnst
import peyes._DataModels.config as cnfg
from peyes._utils.event_utils import parse_label, calculate_num_samples
from peyes._DataModels.Event import BaseEvent, EventSequenceType
from peyes._DataModels.UnparsedEventLabel import UnparsedEventLabelType, UnparsedEventLabelSequenceType
from peyes._DataModels.EventLabelEnum import EventLabelEnum
from peyes._DataModels.Detector import BaseDetector
from peyes._DataModels.Detector import (
    IVTDetector, IVVTDetector, IDTDetector, IDVTDetector, EngbertDetector, NHDetector, REMoDNaVDetector
)


def create_detector(
        algorithm: str,
        missing_value: float,
        min_event_duration: float,
        pad_blinks_time: float,
        name: str = None,
        **kwargs
) -> BaseDetector:
    """
    Creates a gaze event detector with the specified parameters.
    :param algorithm: name of the detection algorithm to use
    :param missing_value: value indicating missing data in the gaze data
    :param min_event_duration: minimum duration of a gaze event (ms)
    :param pad_blinks_time: duration to pad before and after each detected blink event (ms)
    :param name: name of the detector. If None, the detector's class name is used.

    Additional keyword arguments required for detector initialization:
    - IVT:
        :keyword saccade_velocity_threshold: float; velocity threshold to separates saccades from fixations (deg/s)
    - IVVT:
        :keyword saccade_velocity_threshold: float; velocity threshold to separates saccades from smooth pursuits (deg/s)
        :keyword smooth_pursuit_velocity_threshold: float; the velocity threshold that separates smooth pursuits from fixations (deg/s)
    - IDT:
        :keyword dispersion_threshold: dispersion threshold that separates saccades from fixations (visual angle degrees)
        :keyword window_duration: the duration of the window used to calculate the dispersion threshold (ms)
    - IDVT:
        :keyword dispersion_threshold: dispersion threshold that separates fixations from smooth pursuits (visual angle degrees)
        :keyword window_duration: the duration of the window used to calculate the dispersion threshold (ms)
        :keyword saccade_velocity_threshold: float; velocity threshold to separates smooth pursuits from saccades (deg/s)
    - Engbert:
        :keyword lambda_param: float; multiplication coefficient used for calculating saccade threshold
        :keyword deriv_window_size: int; number of samples used to compute axial velocity
    - NH:
        :keyword filter_duration_ms: int; Savitzky-Golay filter's duration (ms), default: 2×min saccade duration
        :keyword filter_polyorder: int; Savitzky-Golay filter's polynomial order, default: 2
        :keyword saccade_max_velocity: maximum saccade velocity (deg/s), default: 1000
        :keyword saccade_max_acceleration: maximum saccade acceleration (deg/s^2), default: 100000
        :keyword min_saccade_duration: minimum saccade duration (ms)
        :keyword min_fixation_duration: minimum fixation duration (ms)
        :keyword max_pso_duration: maximum PSO duration (ms)
        :keyword alpha_param: weight of saccade onset threshold when detecting saccade offset, default: 0.7
        :keyword ignore_short_peak_durations: if True, excludes sporadic occurrences of peak velocity when detecting
            saccade peaks, default: True
        :keyword allow_high_psos: if True, includes "high" PSOs, i.e., PSOs with max velocity exceeding saccades' peak
            threshold but lower than the preceding saccade, default: True
    - REMoDNaV:
        :keyword min_saccade_duration: int; the minimum duration of a saccade (ms)
        :keyword saccade_initial_velocity_threshold: float; the initial velocity threshold for saccade detection (deg/s), default: 300
        :keyword saccade_context_window_duration: int; the duration of the context window for saccade detection (ms), default: 1000
        :keyword saccade_initial_max_freq: float; the initial maximum frequency for saccade detection (Hz), default: 2.0
        :keyword saccade_onset_threshold_noise_factor: float; the noise factor for saccade onset threshold, default: 5.0
        :keyword min_smooth_pursuit_duration: int; the minimum duration of a smooth pursuit (ms)
        :keyword smooth_pursuits_lowpass_cutoff_freq: float; the lowpass cutoff frequency for smooth pursuit detection (Hz), default: 4.0
        :keyword smooth_pursuit_drift_velocity_threshold: float; the drift velocity threshold for smooth pursuit detection (deg/s), default: 2.0
        :keyword min_fixation_duration: int; the minimum duration of a fixation (ms)
        :keyword min_blink_duration: int; the minimum duration of a blink (ms)
        :keyword max_pso_duration: int; the maximum duration of a PSO (ms)
        :keyword savgol_filter_polyorder: int; the polynomial order for the Savitzky-Golay filter, default: 2
        :keyword savgol_filter_duration_ms: int; the duration of the Savitzky-Golay filter (ms), default: 19
        :keyword median_filter_duration_ms: int; the duration of the median filter (ms)
        :keyword max_velocity: float; the maximum velocity of the gaze data (deg/s), default: 1500

    :return: a detector object
    """
    algorithm_lower = algorithm.lower().strip().replace('-', '').removesuffix('detector')
    if algorithm_lower == 'ivt':
        default_params = IVTDetector.get_default_params()
        return IVTDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'ivvt':
        default_params = IVVTDetector.get_default_params()
        return IVVTDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'idt':
        default_params = IDTDetector.get_default_params()
        return IDTDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'idvt':
        default_params = IDVTDetector.get_default_params()
        return IDVTDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'engbert':
        default_params = EngbertDetector.get_default_params()
        return EngbertDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'nh':
        default_params = NHDetector.get_default_params()
        return NHDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    elif algorithm_lower == 'remodnav':
        default_params = REMoDNaVDetector.get_default_params()
        return REMoDNaVDetector(
            missing_value=missing_value,
            min_event_duration=min_event_duration,
            pad_blinks_ms=pad_blinks_time,
            name=name,
            **{k: kwargs.get(k, default_params[k]) for k in default_params.keys()}
        )
    else:
        raise NotImplementedError(f'Detector `{algorithm}` is not implemented.')


def create_events(
        labels: Union[UnparsedEventLabelType, UnparsedEventLabelSequenceType],
        t: np.ndarray,
        x: np.ndarray,
        y: np.ndarray,
        pupil: np.ndarray,
        viewer_distance: float,
        pixel_size: float,
) -> Union[BaseEvent, EventSequenceType]:
    """
    Create gaze-events from the given data.
    If `labels` is a single label, creates a single event spanning the entire data.
    Otherwise, `labels` must have the same length as `t` and events will be created for each "chunk" of subsequent
    identical labels.

    :param labels: The event label(s)
    :param t: timestamps (ms)
    :param x: horizontal gaze position (pixels)
    :param y: vertical gaze position (pixels)
    :param pupil: pupil size (mm)
    :param viewer_distance: distance from the viewer's eyes to the screen (cm)
    :param pixel_size: size of a pixel on the screen (cm)
    :return: the generated event(s)
    """
    if isinstance(labels, UnparsedEventLabelType):
        label = parse_label(labels)
        return BaseEvent.make(label, t, x, y, pupil, viewer_distance, pixel_size)
    return BaseEvent.make_multiple(
        np.array([parse_label(l) for l in labels]), t, x, y, pupil, viewer_distance, pixel_size
    )


def create_boolean_channel(
        channel_type: str,
        data: Union[UnparsedEventLabelSequenceType, EventSequenceType],
        sampling_rate: float = None,
        min_num_samples: int = None,
) -> np.ndarray:
    """
    Converts the given labels or events to a boolean array (MNE-style event channel), indicating either event onsets or
    offsets, depending on the specified channel type.
    Raises a ValueError if neither `labels` and `events` are provided, or if both are provided.

    :param channel_type: either 'start'/'onset' or 'end'/'offset'
    :param data: array-like of event labels or Event objects
    :param sampling_rate: the sampling rate of the recorded data; required if `data` is a series of Event objects
    :param min_num_samples: the number of samples in the output sequence; required if `data` is a series of Event
        objects. If None, the number of samples is determined by the total duration of the provided events.

    :return: array of boolean values, where `True` indicates onsets or offsets
    """
    channel_type_lower = channel_type.lower().strip()
    if channel_type_lower not in {cnst.START_STR, cnst.ONSET_STR, cnst.END_STR, cnst.OFFSET_STR}:
        raise ValueError(f"Invalid channel type: {channel_type}")
    if len(data) == 0:
        # no events or labels provided -> return an empty array
        return np.zeros(np.nanmin(len(data), min_num_samples), dtype=bool)
    if all(isinstance(e, BaseEvent) for e in data):
        # data is of type EventSequenceType
        return _events_to_boolean_channel(channel_type_lower, data, sampling_rate, min_num_samples)
    if all(isinstance(l, UnparsedEventLabelType) for l in data):
        # data is of type UnparsedEventLabelSequenceType
        return _labels_to_boolean_channel(channel_type_lower, data)
    unknown_types = set([
        type(datum) for datum in data if not (isinstance(datum, BaseEvent) or isinstance(datum, UnparsedEventLabelType))
    ])
    raise TypeError(f"Argument `data` contains unknown types: {unknown_types}")


def _labels_to_boolean_channel(
        channel_type: str,
        labels: UnparsedEventLabelSequenceType,
) -> np.ndarray:
    """
    Converts the given labels to a boolean array (MNE-style event channel), indicating either event onsets or offsets.

    :param channel_type: either 'start'/'onset' or 'end'/'offset'
    :param labels: array-like of unparsed event labels

    :return: array of boolean values, where `True` indicates onsets or offsets
    """
    parsed_labels = np.array([parse_label(l) for l in labels])
    bool_channel = np.zeros_like(labels, dtype=bool)
    if channel_type.lower() == cnst.START_STR or channel_type.lower() == cnst.ONSET_STR:
        bool_channel[0] = True
        bool_channel[1:] = np.diff(parsed_labels) != 0
    elif channel_type.lower() == cnst.END_STR or channel_type.lower() == cnst.OFFSET_STR:
        bool_channel[-1] = True
        bool_channel[:-1] = np.diff(parsed_labels) != 0
    else:
        raise ValueError(f"Invalid channel type: {channel_type}")
    bool_channel[parsed_labels == EventLabelEnum.UNDEFINED] = False  # ignore undefined onsets/offsets
    return bool_channel


def _events_to_boolean_channel(
        channel_type: str,
        events: EventSequenceType,
        sampling_rate: float,
        min_num_samples=None,
) -> (np.ndarray, np.ndarray):
    """
    Converts the given events to a boolean array (MNE-style event channel), indicating either event onsets or offsets.

    :param channel_type: either 'start'/'onset' or 'end'/'offset'
    :param events: array-like of Event objects
    :param sampling_rate: the sampling rate of the recorded data
    :param min_num_samples: the number of samples in the output sequence. If None, the number of samples is determined
        by the total duration of the provided events.

    :return: array of boolean values, where `True` indicates the event onset or offset
    """
    global_start_time = min(e.start_time for e in events)
    global_end_time = max(e.end_time for e in events)
    num_samples = calculate_num_samples(global_start_time, global_end_time, sampling_rate, min_num_samples)
    bool_channel = np.zeros(num_samples, dtype=bool)
    for e in events:
        corrected_start_time, corrected_end_time = e.start_time - global_start_time, e.end_time - global_start_time
        if channel_type.lower() == cnst.START_STR or channel_type.lower() == cnst.ONSET_STR:
            start_sample = int(np.round(corrected_start_time * sampling_rate / cnst.MILLISECONDS_PER_SECOND))
            bool_channel[start_sample] = True
        elif channel_type.lower() == cnst.END_STR or channel_type.lower() == cnst.OFFSET_STR:
            end_sample = int(np.round(corrected_end_time * sampling_rate / cnst.MILLISECONDS_PER_SECOND))
            bool_channel[end_sample - 1] = True
        else:
            raise ValueError(f"Invalid channel type: {channel_type}")
    return bool_channel
